from itertools import zip_longest

from elecsim.role.plants.costs.plant_cost_calculation import PlantCostCalculations

"""
File name: non_fuel_lcoe_calculation
Date created: 18/12/2018
Feature: # Calculates the costs of non fuel plants such as LCOE and marginal cost.
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"

class NonFuelCostCalculation(PlantCostCalculations):

    def __init__(self, construction_year, capacity_mw, average_load_factor, pre_dev_period, construction_period, operating_period, pre_dev_spend_years, construction_spend_years, pre_dev_cost_per_mw, construction_cost_per_mw, infrastructure, fixed_o_and_m_per_mw, variable_o_and_m_per_mwh, insurance_cost_per_mw, connection_cost_per_mw, efficiency):
        """
        Power plant of plant_type that does not use plant_type.
        """
        super().__init__(capacity_mw, construction_year, average_load_factor, pre_dev_period, construction_period, operating_period, pre_dev_spend_years, construction_spend_years, pre_dev_cost_per_mw, construction_cost_per_mw, infrastructure, fixed_o_and_m_per_mw, variable_o_and_m_per_mwh, insurance_cost_per_mw, connection_cost_per_mw)


    def calculate_lcoe(self, discount_rate):
        """
        Function which calculates the levelised cost of electricity for this power plant instance
        :return: Returns LCOE value for power plant
        """

        # Calculations to convert into total costs for this power plant instance

        elec_gen, total_costs = self.calculate_total_costs()

        # Discount data
        discounted_total_costs = self._discount_data(total_costs, discount_rate)
        discounted_electricity_generated = self._discount_data(elec_gen, discount_rate)

        # Sum total costs over life time of plant
        discounted_costs_sum = sum(discounted_total_costs)
        discounted_electricity_sum = sum(discounted_electricity_generated)

        lcoe = discounted_costs_sum/discounted_electricity_sum
        return lcoe

    def calculate_total_costs(self):
        capex = self._capex()
        opex = self._opex_cost()
        elec_gen = self._electricity_generated()
        total_costs = self._total_costs(capex, opex)
        return elec_gen, total_costs

    def calculate_short_run_marginal_cost(self, model):
        return self.variable_o_and_m_per_mwh


    def _total_costs(self, capex, opex):
        """
        Calculates total costs of plant by adding capital expenses plus operating expenses.
        :return: Total costs over lifetime of plant
        """

        total_costs = [x + y for x, y in zip_longest(capex, opex, fillvalue=0)]
        return total_costs





from functools import lru_cache
from inspect import signature
from itertools import zip_longest
from logging import getLogger

import pandas as pd
from numpy import npv, pmt, divide

import numpy as np
np.seterr(divide='ignore', invalid='ignore')

from elecsim.plants.fuel.capacity_factor.capacity_factor_calculations import get_capacity_factor
from elecsim.plants.plant_costs.estimate_costs.estimate_costs import create_power_plant
from elecsim.plants.plant_type.fuel_plant import FuelPlant
from elecsim.role.investment.predict_load_duration_prices import get_price_duration_curve
from elecsim.role.market.latest_market_data import LatestMarketData
from elecsim.role.plants.costs.fuel_plant_cost_calculations import FuelPlantCostCalculations
# from elecsim.scen_error.scenario_data import modern_plant_costs
# from elecsim.scen_error.scenario_data import nuclear_wacc, non_nuclear_wacc
import elecsim.scenario.scenario_data
from functools import lru_cache

logger = getLogger(__name__)

"""
File name: calculate_npv
Date created: 04/01/2019
Feature: # Contains functionality to assess options of investment and return lowest NPV for decision to be made.
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"


class CalculateNPV:
    def __init__(self, model, difference_in_discount_rate, look_back_years):
        self.model = model
        self.difference_in_discount_rate = difference_in_discount_rate
        self.look_back_years = look_back_years

    def get_positive_npv_plants_list(self):
        npv_rows = self.get_positive_npv_plants()
        logger.debug("npv_rows: {}".format(npv_rows))
        result = [(individual_plant_data.capacity, individual_plant_data.plant_type) for individual_plant_data in
                  npv_rows.itertuples()]
        return result

    def get_positive_npv_plants(self):
        npv_data = self.compare_npv()
        logger.debug("predicted npv data: \n {}".format(npv_data))

        npv_positive = npv_data[npv_data.npv_per_mw > 0]
        return npv_positive

<<<<<<< HEAD
    def compare_npv(self):
=======
    def  compare_npv(self):
>>>>>>> 5f3c373861c398621fed06ebb3fe989ceb1733a9
        cost_list = []

        for plant_type in ['CCGT', 'Coal', 'Nuclear', 'Onshore', 'Offshore', 'PV', 'Recip_gas', 'Recip_diesel']:
            plant_cost_data = elecsim.scenario.scenario_data.modern_plant_costs[(elecsim.scenario.scenario_data.modern_plant_costs.Type == plant_type) & (elecsim.scenario.scenario_data.modern_plant_costs.Plant_Size>5)]
            for plant_row in plant_cost_data.itertuples():

                npv = self.calculate_npv(plant_row.Type, plant_row.Plant_Size)
                dict = {"npv_per_mw": npv, "capacity": plant_row.Plant_Size, "plant_type": plant_row.Type}
                cost_list.append(dict)

        npv_results = pd.DataFrame(cost_list)

        sorted_npv = npv_results.sort_values(by='npv_per_mw', ascending=False)
        logger.debug("sorted_npv: \n {}".format(sorted_npv))
        return sorted_npv

    # @lru_cache(maxsize=10000)
    def calculate_npv(self, plant_type, plant_size):
        # Forecast segment prices

<<<<<<< HEAD
=======

>>>>>>> 5f3c373861c398621fed06ebb3fe989ceb1733a9
        forecasted_segment_prices = self._get_price_duration_predictions()
        # logger.info("Forecasted price duration curve: {}".format(forecasted_segment_prices))

        power_plant = create_power_plant("PowerPlantName", self.model.year_number, plant_type, plant_size)

        # Forecast marginal costs
        short_run_marginal_cost = self._get_predicted_marginal_cost(power_plant)
        logger.debug("short run marginal cost: {}".format(short_run_marginal_cost))

        forecasted_segment_prices = self._clean_segment_prices(forecasted_segment_prices)
        logger.debug("forecasted_segment_prices: \n {}".format(forecasted_segment_prices))

        self._get_profit_per_mwh(forecasted_segment_prices, short_run_marginal_cost)

        self._get_profit_per_segment(forecasted_segment_prices, power_plant=power_plant)

        self._get_total_hours_to_run(forecasted_segment_prices, power_plant)

        self._get_total_yearly_income(forecasted_segment_prices, power_plant)

        logger.debug("total_hours_predicted_to_run: \n {}".format(forecasted_segment_prices))

        # total_profit_for_year = sum(forecasted_segment_prices['_total_profit_per_segment'])
        total_running_hours = sum(forecasted_segment_prices['_total_running_hours'])
        total_yearly_income = sum(forecasted_segment_prices['total_income'])
        logger.debug("total_yearly_income: {}".format(total_yearly_income))
        yearly_capital_cost = self._get_capital_outflow(power_plant)

        logger.debug("yearly_capital_cost: {}".format(yearly_capital_cost))

        yearly_operating_cash_flow = self._calculate_yearly_operation_cashflow(power_plant, short_run_marginal_cost,
                                                                               total_running_hours,
                                                                               total_yearly_income, yearly_capital_cost)

        total_cash_flow = yearly_capital_cost + yearly_operating_cash_flow

        logger.debug('total_cash_flow: {}'.format(total_cash_flow))

        if power_plant.plant_type == "Nuclear":
            self.weighted_average_cost_capital = elecsim.scenario.scenario_data.nuclear_wacc + self.difference_in_discount_rate
        else:
            self.weighted_average_cost_capital = elecsim.scenario.scenario_data.non_nuclear_wacc + self.difference_in_discount_rate

        # logger.debug("cash_flow_wacc: {}".format(cash_flow_wacc))
        logger.debug("discount rate: {}".format(self.weighted_average_cost_capital))

        npv_power_plant = npv(self.weighted_average_cost_capital, total_cash_flow)

        logger.debug("npv_power_plant: {}".format(npv_power_plant))

        NPVp = divide(npv_power_plant, (power_plant.capacity_mw * total_running_hours))
        return NPVp

    def _get_yearly_profit_per_mwh(self, power_plant, total_running_hours, yearly_cash_flow):
        yearly_cash_flow_per_mwh = yearly_cash_flow / (total_running_hours * power_plant.capacity_mw)
        return yearly_cash_flow_per_mwh

    def _calculate_yearly_operation_cashflow(self, power_plant, short_run_marginal_cost, total_running_hours,
                                             total_yearly_income,
                                             yearly_capital_cost):
        # logger.debug("total_profit_for_year: {}, total running hours: {}".format(total_profit_for_year, _total_running_hours))
        operational_costs = short_run_marginal_cost * total_running_hours * power_plant.capacity_mw
        # logger.debug("yearly_capital_cost: {}".format(yearly_capital_cost))
        # logger.debug("total yearly cost: {}, total yearly _income: {}".format(operational_costs, total_yearly_income))
        result = [total_yearly_income - operational_costs] * int(power_plant.operating_period)
        return result

    def _get_capital_outflow(self, power_plant):
        power_plant_vars = vars(power_plant)
        logger.debug("power_plant_vars: {}".format(power_plant_vars))
        func = FuelPlantCostCalculations
        vars_required = signature(func)._parameters
        logger.debug("vars_required: {}".format(vars_required))
        power_plant_vars = {key: value for key, value in power_plant_vars.items() if key in vars_required}
        # yearly_outflow = FuelPlantCostCalculations(**power_plant_vars).calculate_yearly_outflow()

        capital_costs = FuelPlantCostCalculations(**power_plant_vars)._capex()
        logger.debug("capital_cost_outflow: {}".format(capital_costs))
        capital_costs_outflow = [-x for x in capital_costs]
        return capital_costs_outflow

    def _get_total_yearly_income(self, forecasted_segment_prices, power_plant):
        # forecasted_segment_prices['total_income'] = forecasted_segment_prices.apply(
        #     lambda x: self._income(x, power_plant.capacity_mw), axis=1)
        vectorized_income = np.vectorize(self._income)
        forecasted_segment_prices['total_income'] = vectorized_income(forecasted_segment_prices['predicted_profit_per_mwh'].values, forecasted_segment_prices['_total_running_hours'].values, forecasted_segment_prices['accepted_price'].values, power_plant.capacity_mw)


    def _get_total_hours_to_run(self, forecasted_segment_prices, power_plant):
        # forecasted_segment_prices['_total_running_hours'] = forecasted_segment_prices.apply(
        #     lambda x: self._total_running_hours(x, power_plant), axis=1)

        if isinstance(power_plant, FuelPlant):
            vectorized_total_running_hours_fuel_plant = np.vectorize(self._total_running_hours_fuel_plant)

            forecasted_segment_prices['_total_running_hours'] = vectorized_total_running_hours_fuel_plant(forecasted_segment_prices['predicted_profit_per_mwh'].values, forecasted_segment_prices['num_of_hours'].values)
        else:
            vectorized_total_running_hours_renewable_plant = np.vectorize(self._total_running_hours_renewable_plant)

            forecasted_segment_prices['_total_running_hours'] = vectorized_total_running_hours_renewable_plant(forecasted_segment_prices['predicted_profit_per_mwh'].values, forecasted_segment_prices['num_of_hours'].values, forecasted_segment_prices['segment_hour'].values, power_plant)



    def _get_profit_per_segment(self, forecasted_segment_prices, power_plant):
        # forecasted_segment_prices['_total_profit_per_segment'] = forecasted_segment_prices.apply(
        #     lambda x: self._total_profit_per_segment(x, power_plant.capacity_mw), axis=1)

        vectorized_total_profit_per_segment = np.vectorize(self._total_profit_per_segment)

        forecasted_segment_prices['_total_profit_per_segment'] = vectorized_total_profit_per_segment(forecasted_segment_prices['predicted_profit_per_mwh'].values, forecasted_segment_prices['num_of_hours'].values, power_plant.capacity_mw)

    def _get_profit_per_mwh(self, forecasted_segment_prices, short_run_marginal_cost):
        forecasted_segment_prices['predicted_profit_per_mwh'] = forecasted_segment_prices[
                                                                    'accepted_price'] - short_run_marginal_cost

    def _clean_segment_prices(self, forecasted_segment_prices):
        # forecasted_segment_prices = forecasted_segment_prices.to_frame().reset_index()
        forecasted_segment_prices = forecasted_segment_prices.reset_index()

        forecasted_segment_prices['num_of_hours'] = abs(forecasted_segment_prices.segment_hour.diff())
        forecasted_segment_prices = forecasted_segment_prices.dropna()
        return forecasted_segment_prices

    def _get_predicted_marginal_cost(self, power_plant):
        market_data = LatestMarketData(model=self.model)
        short_run_marginal_cost = market_data.get_predicted_marginal_cost(power_plant, self.look_back_years)
        return short_run_marginal_cost

    def _get_price_duration_predictions(self):
        predicted_price_duration_curve = get_price_duration_curve(self.model, self.look_back_years)

        return predicted_price_duration_curve

    # @staticmethod
    # def _total_profit_per_segment(row, capacity):
    #     if row['predicted_profit_per_mwh'] > 0:
    #         total_profit = row['num_of_hours'] * row['predicted_profit_per_mwh'] * capacity
    #     else:
    #         total_profit = 0
    #     return total_profit

    @staticmethod
    def _total_profit_per_segment(predicted_profit_per_mwh, num_of_hours, capacity):
        if predicted_profit_per_mwh > 0:
            return num_of_hours * predicted_profit_per_mwh * capacity
        else:
            return 0








    # def _total_running_hours(self, row, power_plant):
    #     if isinstance(power_plant, FuelPlant):
    #         if row['predicted_profit_per_mwh'] > 0:
    #             running_hours = row['num_of_hours']
    #         else:
    #             running_hours = 0
    #     else:
    #         if row['predicted_profit_per_mwh'] > 0:
    #             capacity_factor = get_capacity_factor(self.model.market_time_splices, power_plant.plant_type, row.segment_hour)
    #             running_hours = capacity_factor * row['num_of_hours']
    #         else:
    #             running_hours = 0
    #     return running_hours
    #

    def _total_running_hours_fuel_plant(self, predicted_profit_per_mwh, num_of_hours):

        if predicted_profit_per_mwh > 0:
            return num_of_hours
        else:
            return 0

    def _total_running_hours_renewable_plant(self, predicted_profit_per_mwh, num_of_hours, segment_hour, power_plant):
        if predicted_profit_per_mwh > 0:
            capacity_factor = get_capacity_factor(self.model.market_time_splices, power_plant.plant_type, segment_hour)
            return capacity_factor * num_of_hours
        else:
            return 0



    # @staticmethod
    # def _income(row, capacity):
    #     if row['predicted_profit_per_mwh'] > 0:
    #         running_hours = row['_total_running_hours'] * row['accepted_price'] * capacity
    #     else:
    #         running_hours = 0
    #     return running_hours

    @staticmethod
    def _income(predicted_profit_per_mwh, _total_running_hours, accepted_price, capacity):

        if predicted_profit_per_mwh > 0:
            return _total_running_hours * accepted_price * capacity
        else:
            return 0

        #
        # if row['predicted_profit_per_mwh'] > 0:
        #     running_hours = row['_total_running_hours'] * row['accepted_price'] * capacity
        # else:
        #     running_hours = 0
        # return running_hours


<<<<<<< HEAD
=======
# @lru_cache(maxsize=1024)
>>>>>>> 5f3c373861c398621fed06ebb3fe989ceb1733a9
def get_most_profitable_plants_by_npv(model, difference_in_discount_rate, look_back_period):
    npv_calculation = CalculateNPV(model, difference_in_discount_rate, look_back_period)

    potential_plant_data = npv_calculation.get_positive_npv_plants_list()

    return potential_plant_data


def get_yearly_payment(power_plant, interest_rate, down_payment_perc):
    loan_percentage = 1-down_payment_perc
    pre_dev_downpayments = [pre_dev_year * power_plant.pre_dev_cost_per_mw * power_plant.capacity_mw * loan_percentage for pre_dev_year in
                            power_plant.pre_dev_spend_years]

    pre_dev_years_to_pay_back = [
        12 * (power_plant.operating_period + power_plant.construction_period + downpayment_date) for
        downpayment_date in list(reversed(range(1, len(pre_dev_downpayments) + 1)))]

    # logger.debug("pre_dev_years_to_pay_back: {}".format(pre_dev_years_to_pay_back))

    pre_dev_interest = [interest_rate / 12] * int(power_plant.pre_dev_period)

    # logger.debug("pre_dev_interest: {}".format(pre_dev_interest))

    construction_downpayments = [construction_year * power_plant.construction_cost_per_mw * power_plant.capacity_mw * loan_percentage for
                                 construction_year in power_plant.construction_spend_years]

    construction_interest = [interest_rate / 12] * int(power_plant.construction_period)

    construction_years_to_pay_back = [12 * (power_plant.operating_period + downpayment_date) for
                                      downpayment_date in list(reversed(range(1, len(construction_downpayments) + 1)))]

    pre_dev_repayments = pmt(rate=pre_dev_interest, nper=pre_dev_years_to_pay_back, pv=pre_dev_downpayments, fv=0,
                             when=0)

    total_years_to_pay_pre_dev = [power_plant.operating_period + power_plant.construction_period + downpayment_date for
                          downpayment_date in list(reversed(range(1, len(pre_dev_downpayments) + 1)))]

    pre_dev_lumpsums = []
    for years_to_pay, pre_dev_downpayment in zip(total_years_to_pay_pre_dev, pre_dev_repayments):
        # logger.debug('years_to_pay: {}'.format(years_to_pay))
        pre_dev_payback = [pre_dev_downpayment] * int(years_to_pay)
        pre_dev_lumpsums.append(pre_dev_payback)

    # logger.debug("pre_dev_lumpsums: {}".format(pre_dev_lumpsums))
    # logger.debug("len of lumpsums: {}, len of lumpsums[0]: {}, len of lumpsums[1]: {}, len of lumpsums[2]: {}".format(len(pre_dev_lumpsums), len(pre_dev_lumpsums[0]),len(pre_dev_lumpsums[1]), len(pre_dev_lumpsums[2])))

    construction_repayments = pmt(rate=construction_interest, nper=construction_years_to_pay_back,
                                  pv=construction_downpayments, fv=0, when=0)

    construction_lumpsums = []
    total_years_to_pay_construction = [(power_plant.operating_period + downpayment_date) for downpayment_date in list(reversed(range(1, len(construction_downpayments) + 1)))]
    for years_to_pay, construction_downpayment in zip(total_years_to_pay_construction, construction_repayments):
        construction_payback = [construction_downpayment] * int(years_to_pay)
        construction_lumpsums.append(construction_payback)


    total_pre_dev_payments = [sum(i) for i in zip_longest(*reversed(pre_dev_lumpsums), fillvalue=0)][::-1]
    # logger.debug("total_pre_dev_payments:{}".format(total_pre_dev_payments))

    # Join construction and pre-dev payments together
    total_construction_payments = [sum(i) for i in zip_longest(*reversed(construction_lumpsums), fillvalue=0)][::-1]

    # Add construction payments and pre-development payments together
    total_payments = [a+b for a, b in zip_longest(reversed(total_pre_dev_payments), reversed(total_construction_payments), fillvalue=0)][::-1]

    # repayments = pre_dev_repayments + construction_repayments

    return total_payments


def select_yearly_payback_payment_for_year(power_plant, interest, downpayment_percentage, model):
    total_payments = get_yearly_payment(power_plant, interest, downpayment_percentage)
    years_since_construction = model.year_number-power_plant.construction_year
    if years_since_construction > power_plant.operating_period+power_plant.construction_period+power_plant.pre_dev_period:
        payment_in_current_year=0
    else:
        try:
            payment_in_current_year = total_payments[years_since_construction]
        except IndexError:
            payment_in_current_year=0
    return payment_in_current_year



import datetime as dt
import logging
import os
from random import uniform, randint, sample
from time import perf_counter
import importlib.util
import time
import dropbox

import numpy as np
import pandas as pd
from mesa import Model
from mesa.datacollection import DataCollector
import pickle
from ray.rllib.utils.policy_client import PolicyClient

from elecsim.plants.fuel.capacity_factor.capacity_factor_calculations import get_capacity_factor

from elecsim.agents.demand.demand import Demand
from elecsim.agents.demand.multi_day_demand import MultiDayDemand
from elecsim.agents.generation_company.gen_co import GenCo
from elecsim.constants import ROOT_DIR
from elecsim.market.electricity.market.power_exchange import PowerExchange
from elecsim.mesa_addons.scheduler_addon import OrderedActivation
from elecsim.plants.plant_costs.estimate_costs.estimate_costs import create_power_plant
from elecsim.plants.plant_type.fuel_plant import FuelPlant
import elecsim.data_manipulation.data_modifications.scenario_modifier as scen_mod
from elecsim.role.market.latest_market_data import LatestMarketData

import elecsim.scenario.scenario_data

import os

logger = logging.getLogger(__name__)


"""Model.py: Model for the electricity landscape world"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "Alexander@Kell.es"

class World(Model):
    """
    Model for the electricity landscape world
    """

    def __init__(self, initialization_year, scenario_file=None, fitting_params=None, long_term_fitting_params=None, future_price_uncertainty_m = None, future_price_uncertainty_c = None, carbon_price_scenario=None, demand_change=None, demand_distribution=None, number_of_steps=32, total_demand=None, number_of_agents=None, market_time_splices=1, data_folder=None, time_run=False, nuclear_subsidy=None, highest_demand=None, log_level="warning", client_rl=None, distribution_name = None):
        """
        Initialize an electricity market in a particular country. Provides the ability to change scenarios from this constructor.
        :param int initialization_year: Year to begin simulation.
        :param list: float carbon_price_scenario: Scenario containing carbon price for each year of simulation.
        :param list: float demand_change: Scenario containing change in demand between each year of simulation.
        :param int number_of_steps: Total number of years to run scenario.
        :param int total_demand: Total size of country's demand.
        :param str data_folder: Directory and folder to save run data to
        :param bool time_run:
        :param str log_level:
        """

        self.start = perf_counter()
        logger.info("start: {}".format(self.start))
        # Set up model objects
        self.year_number = initialization_year
        self.years_from_start = 0
        self.step_number = 0
        self.unique_id_generator = 0
        self.time_run = time_run
        self.max_number_of_steps = number_of_steps
        self.average_electricity_price = 0
        self.market_time_splices = market_time_splices

        self.nuclear_subsidy = nuclear_subsidy

        self.set_log_level(log_level)

        self.overwrite_scenario_file(scenario_file)

        self.override_highest_demand(highest_demand)
        self.override_carbon_scenario(carbon_price_scenario)
        self.override_demand_change(demand_change)

        self.demand_distribution_uncertainty = demand_distribution
        self.distribution_name = distribution_name

        self.override_total_demand(total_demand, number_of_agents)

        self.schedule = OrderedActivation(self)

        # Import company data including financials and plant data
        plant_data = elecsim.scenario.scenario_data.power_plants
        financial_data = elecsim.scenario.scenario_data.company_financials

        # Initialize generation companies using financial and plant data
        self.initialize_gencos(financial_data, plant_data)

        self.last_added_plant = None
        self.last_added_plant_bids = None

        # Create PowerExchange

        if self.market_time_splices == 1:
            self.PowerExchange = PowerExchange(self, demand_distribution)
            self.demand = Demand(self, self.unique_id_generator, elecsim.scenario.scenario_data.segment_time, elecsim.scenario.scenario_data.segment_demand_diff)
        elif self.market_time_splices > 1:
            self.PowerExchange = PowerExchange(self, demand_distribution)
            # self.PowerExchange = HighTemporalExchange(self)
            self.demand = MultiDayDemand(self, self.unique_id_generator, elecsim.scenario.scenario_data.multi_year_data)
        else:
            raise ValueError("market_time_splices must be equal to or larger than 1.")


        self.running = True
        self.beginning_of_year = False

        self.continue_investing = 0
        self.over_invested = False

        self.unique_id_generator += 1
        self.schedule.add(self.demand)
        self.create_data_loggers(data_folder)

        if elecsim.scenario.scenario_data.investment_mechanism == "RL":
            # self.client = PolicyClient("http://rllibserver:9900")
            self.client = client_rl
            self.eid = self.client.start_episode(training_enabled=True)
            self.intial_obs = LatestMarketData(self).get_RL_investment_observations()
            # logger.info("self.intial_obs: {}".format(self.intial_obs))
        elif elecsim.scenario.scenario_data.investment_mechanism == "future_price_fit":
            self.future_price_uncertainty_m = future_price_uncertainty_m
            self.future_price_uncertainty_c = future_price_uncertainty_c
            if fitting_params is not None:
                self.fitting_params = fitting_params
            elif long_term_fitting_params is not None:
                self.long_term_fitting_params = long_term_fitting_params
                self.fitting_params = None
            else:
                raise ValueError("If using future_price_fit you must enter a value for long_term_fitting_params or fitting_params in the constructor of World")


    def step(self, carbon_price=None):
        '''Advance model by one step'''
        self.beginning_of_year = False

        if self.step_number % self.market_time_splices == 0:
            self.start = time.perf_counter()
            self.operate_constructed_plants()
            if self.step_number != 0:
                self.year_number += 1
                self.years_from_start += 1
                # self.operate_constructed_plants()
                self.beginning_of_year = True
                # logger.info("year: {}".format(self.year_number))
                print("{}:".format(self.year_number), end='', flush=True)
            else:
                print("{}:".format(self.year_number), end='', flush=True)

        obs = self.schedule.step()
        self.operate_constructed_plants()

        if self.over_invested:
            return self.datacollector.get_model_vars_dataframe(), self.over_invested

        self.continue_investing = 0
        if carbon_price is not None:
            elecsim.scenario.scenario_data.carbon_price_scenario[self.year_number + 1] = carbon_price
        else:
            elecsim.scenario.scenario_data.carbon_price_scenario = elecsim.scenario.scenario_data.carbon_price_scenario

        if self.beginning_of_year:
            self.dismantle_old_plants()
            self.dismantle_unprofitable_plants()

        self.average_electricity_price = self.PowerExchange.tender_bids(self.demand.segment_hours, self.demand.segment_consumption).accepted_price.mean()
        self.PowerExchange.price_duration_curve = []

        carbon_emitted = self.get_carbon_emitted(self)
        self.settle_gencos_financials()

        self.datacollector.collect(self)
        self.delete_old_bids()

        self.step_number += 1
        print(".", end='', flush=True)

        self.write_scenario_data()

        if isinstance(self.average_electricity_price, np.ndarray):
            self.average_electricity_price = self.average_electricity_price[0]

        if self.step_number % self.market_time_splices == 0:
            end = time.perf_counter()
            print("time taken: {}".format(end-self.start))
            # get_capacity_factor.cache_clear()

        if self.step_number == self.max_number_of_steps and elecsim.scenario.scenario_data.investment_mechanism == "RL":
            obs = LatestMarketData(self).get_RL_investment_observations()
            self.client.end_episode(self.eid, observation=obs)
            del self.client

        logger.debug(self.datacollector.get_model_vars_dataframe())
        return abs(self.average_electricity_price), abs(carbon_emitted)
        # return self.datacollector.get_model_vars_dataframe(), self.over_invested

    def initialize_gencos(self, financial_data, plant_data):
        """
        Creates generation company agents based on financial data and power plants owned. Estimates cost parameters
         of each power plant if data not for power plant not available.
        :param financial_data: Data containing information about generation company's financial status
        :param plant_data: Data containing information about generation company's plants owned, start year and name.
        """

        financial_data = pd.merge(financial_data, plant_data, on="Company", how="inner")
        financial_data = financial_data[['Company', 'cash_in_bank']]

        # Initialising generator company data
        financial_data.cash_in_bank = financial_data.cash_in_bank.replace("nan", np.nan)
        financial_data.cash_in_bank = financial_data.cash_in_bank.fillna(0)
        companies_groups = plant_data.groupby('Company')
        company_financials = financial_data.groupby('Company')

        logger.info("Initialising generation companies with their power plants.")

        # Initialize generation companies with their respective power plants
        for gen_id, ((name, data), (_, financials)) in enumerate(zip(companies_groups, company_financials), 0):
            if financials.Company.iloc[0] != name:
                raise ValueError("Company financials name ({}) and agent name ({}) do not match.".format(financials.Company.iloc[0], name))

            gen_co = GenCo(unique_id=gen_id, model=self, difference_in_discount_rate=round(uniform(-0.03, 0.03), 3), look_back_period=randint(3, 7), name=name, money=financials.cash_in_bank.iloc[0])
            self.unique_id_generator += 1
            # Add power plants to generation company portfolio
            # parent_directory = os.path.dirname(os.getcwd())
            pickle_directory = "{}/../elecsim/data/processed/pickled_data/power_plants/".format(ROOT_DIR)
            for plant in data.itertuples():
                try:
                    power_plant = pickle.load(open("{}{}-{}.pickle".format(pickle_directory, plant.Name, plant.Start_date), "rb"))
                except (OSError, IOError, FileNotFoundError) as e:
                    logger.info("plant: {}".format(plant))
                    power_plant = create_power_plant(plant.Name, plant.Start_date, plant.Simplified_Type, plant.Capacity)
                    pickle.dump(power_plant, open("{}{}-{}.pickle".format(pickle_directory, plant.Name, plant.Start_date), "wb"))
                gen_co.plants.append(power_plant)
            logger.debug('Adding generation company: {}'.format(gen_co.name))
            self.schedule.add(gen_co)
        logger.info("Added generation companies.")


    def get_running_plants(self, plants):
        for plant in plants:
            # if plant.name in elecsim.scenario.scenario_data.known_plant_retirements:

            if plant.construction_year <= 1990 and plant.name != "invested_plant" and plant.name not in elecsim.scenario.scenario_data.known_plant_retirements:
                # Reset old plants that have been modernised with new construction year
                plant.construction_year = randint(self.year_number-15, self.year_number)
                # yield plant
            elif plant.name in elecsim.scenario.scenario_data.known_plant_retirements:
                plant.construction_year = elecsim.scenario.scenario_data.known_plant_retirements[plant.name] - (plant.operating_period + plant.construction_period + plant.pre_dev_period) - 1
                logger.info("plant.name: {}, plant.construction_year: {}".format(plant.name, plant.construction_year))

            if plant.construction_year + plant.operating_period + plant.construction_period + plant.pre_dev_period >= self.year_number:
                yield plant
            else:
                logger.debug("Taking the plant '{}' out of service, year of construction: {}".format(plant.name,
                                                                        plant.construction_year))



    def dismantle_old_plants(self):
        """
        Remove plants that are past their lifetime agent from each agent from their plant list
        """

        gencos = self.get_gencos()

        for genco in gencos:
            plants_filtered = list(self.get_running_plants(genco.plants))
            genco.plants = plants_filtered

    def dismantle_unprofitable_plants(self):

        gencos = self.get_gencos()

        for genco in gencos:
            profitable_plants = list(self.filter_plants_with_no_income(genco.plants))
            genco.plants = profitable_plants

    def filter_plants_with_no_income(self, plants):
            for plant in plants:
                if (self.year_number > 7) and (plant.get_year_of_operation() + 7 < self.year_number):
                    historic_bids = plant.historical_bids
                    # logger.info("historic_bids {}".format(historic_bids))
                    # years_to_look_into = list(range(self.year_number,self.year_number-7,-1))

                    seven_years_previous = self.year_number-7
                    if historic_bids:
                        if historic_bids[-1].year_of_bid > seven_years_previous:
                            yield plant
                        else:
                            logger.debug("Plant {}, type {} is unprofitable. Last accepted bid: {}".format(plant.name, plant.plant_type, historic_bids[-1].year_of_bid))
                            for bid in plant.accepted_bids:
                                del bid
                    else:
                        logger.debug("Plant {}, type {} is unprofitable.".format(plant.name, plant.plant_type))
                        for bid in plant.accepted_bids:
                            del bid
                    # bids_to_check = list(filter(lambda x: x.year_of_bid in years_to_look_into, historic_bids))
                    # total_income_in_previous_years = sum(bid.price_per_mwh for bid in bids_to_check)
                    # for bids in reversed(historic_bids):
                    #     logger.info("bids.year_of_bid: {}".format(bids.year_of_bid))

                    # if total_income_in_previous_years > 0:
                    #     yield plant
                    # else:
                    #     logger.debug("Taking plant: {} out of service.".format(plant.name))
                else:
                    yield plant

    def get_profitable_plants(self, plants):
        for plant in plants:
            if self.step_number > 7 and plant.get_year_of_operation() + 7 > self.year_number:
                historic_bids = plant.historical_bids
                pass

    def operate_constructed_plants(self, minimum_operation_year=2018):
        gencos = self.get_gencos()
        logger.debug("gencos: {}".format(gencos))
        for genco in gencos:
            logger.debug("genco plants: {}".format(genco.plants))
            for plant in genco.plants:
                # logger.debug("plant: {}, year_number: {}, construction year+constructioon_period+predev: {}".format(plant, self.year_number, plant.construction_year + plant.construction_period + plant.pre_dev_period))
                if plant.construction_year <= minimum_operation_year:
                    plant.is_operating = True
                elif (plant.is_operating is False) and (self.year_number >= plant.construction_year + plant.construction_period + plant.pre_dev_period):
                    plant.is_operating = True

    def overwrite_scenario_file(self, scenario_file):
        if scenario_file:
            split_directory = scenario_file.split("/")
            file_name = split_directory[-1]
            spec = importlib.util.spec_from_file_location(file_name, scenario_file)
            scenario_import = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(scenario_import)
            scen_mod.overwrite_scenario_file(scenario_import)

    def settle_gencos_financials(self):
        gencos = self.get_gencos()
        for genco in gencos:
            genco.settle_accounts()
            # genco.delete_old_bids()

    def delete_old_bids(self):
        gencos = self.get_gencos()
        for genco in gencos:
            genco.delete_old_bids()

    def get_gencos(self):
        gencos = [genco for genco in self.schedule.agents if isinstance(genco, GenCo)]
        return gencos

    def clear_all_bids(self):
        gencos = self.get_gencos()
        for genco in gencos:
            genco.delete_old_bids()

    @staticmethod
    def get_capacity_of_plants(model, plant_type):
        gencos = model.get_gencos()
        plants = [plant for genco in gencos for plant in genco.plants if plant.plant_type == plant_type and plant.is_operating]
        total_capacity = sum(plant.capacity_mw for plant in plants)

        return total_capacity


    @staticmethod
    def get_all_plants(model):
        gencos = model.get_gencos()
        plants = [plant for genco in gencos for plant in genco.plants]
        return plants


    @staticmethod
    def get_current_carbon_tax(model):
        carbon_tax = elecsim.scenario.scenario_data.carbon_price_scenario[model.years_from_start]
        return carbon_tax

    @staticmethod
    def get_genco_wealth(model):
        gencos = model.get_gencos()
        total_wealth = 0
        for genco in gencos:
            total_wealth += genco.money
        return total_wealth

    @staticmethod
    def get_electricity_cost(model):
        return model.average_electricity_price

    @staticmethod
    def get_carbon_emitted(model):
        gencos = model.get_gencos()
        bids = World.get_accepted_bids(gencos, FuelPlant)
        
        carbon_emitted = sum(bid.capacity_bid * bid.plant.fuel.co2_density for bid in bids if isinstance(bid.plant, FuelPlant))
        return carbon_emitted

    @staticmethod
    def get_accepted_bid_capacity(model, plant_type):
        gencos = model.get_gencos()
        plants = [plant for genco in gencos for plant in genco.plants if plant.plant_type == plant_type and plant.is_operating]
        capacity_contributed = sum(bid.capacity_bid for plant in plants for bid in plant.accepted_bids)
        return capacity_contributed

    @staticmethod
    def get_accepted_bid_capacity_per_segment_hour(model):
        gencos = model.get_gencos()
        plants = [plant for genco in gencos for plant in genco.plants if plant.is_operating]
        # capacity_contributed = [ if bid.segment_hours==]
        bids_dataframe = [bid.to_dict() for plant in plants for bid in plant.accepted_bids]
        return bids_dataframe


    @staticmethod
    def get_accepted_bids(gencos, plant_type=None):
        if plant_type:
            bids = list(
                accepted_bids for genco in gencos for plants in genco.plants for accepted_bids in plants.accepted_bids if
                isinstance(plants, plant_type))
        else:
            bids = list(
                accepted_bids for genco in gencos for plants in genco.plants for accepted_bids in plants.accepted_bids)

        return bids

    def stratify_data(self, demand):
        power_plants = elecsim.scenario.scenario_data.power_plants
        frac_to_scale = demand/power_plants.Capacity.sum()

        stratified_sample = power_plants.groupby(['Fuel'], as_index=False).apply(lambda x: x.sample(frac=frac_to_scale, replace=True))
        return stratified_sample




    def set_log_level(self, log_level):
        if log_level.lower() == "warning":
            logging.basicConfig(level=logging.WARNING)
        elif log_level.lower() == "info":
            logging.basicConfig(level=logging.INFO)
        elif log_level.lower() == "debug":
            logging.basicConfig(level=logging.DEBUG)
        else:
            raise ValueError("log_level must be warning, info or debug and not {}".format(log_level))

    def create_data_loggers(self, data_folder):
        self.data_folder = data_folder
        self.datacollector = DataCollector(
            model_reporters={"contributed_CCGT": lambda m: self.get_accepted_bid_capacity(m, "CCGT"),
                             "contributed_Coal": lambda m: self.get_accepted_bid_capacity(m, "Coal"),
                             "contributed_Onshore": lambda m: self.get_accepted_bid_capacity(m, "Onshore"),
                             "contributed_Offshore": lambda m: self.get_accepted_bid_capacity(m, "Offshore"),
                             "contributed_PV": lambda m: self.get_accepted_bid_capacity(m, "PV"),
                             "contributed_Nuclear": lambda m: self.get_accepted_bid_capacity(m, "Nuclear"),
                             "contributed_Recip_gas": lambda m: self.get_accepted_bid_capacity(m, "Recip_gas"),
                             "contributed_Biomass": lambda m: self.get_accepted_bid_capacity(m, "Biomass"),

                             # "hourly_accepted_bids": lambda m: self.get_accepted_bid_capacity_per_segment_hour(m),

                             "total_CCGT": lambda m: self.get_capacity_of_plants(m, "CCGT"),
                             "total_Coal": lambda m: self.get_capacity_of_plants(m, "Coal"),
                             "total_Onshore": lambda m: self.get_capacity_of_plants(m, "Onshore"),
                             "total_Offshore": lambda m: self.get_capacity_of_plants(m, "Offshore"),
                             "total_PV": lambda m: self.get_capacity_of_plants(m, "PV"),
                             "total_Nuclear": lambda m: self.get_capacity_of_plants(m, "Nuclear"),
                             "total_Recip_gas": lambda m: self.get_capacity_of_plants(m, "Recip_gas"),
                             "Carbon_tax": lambda m: self.get_current_carbon_tax(m),
                             "total_genco_wealth": lambda m: self.get_genco_wealth(m),
                             "Electricity_cost": lambda m: self.get_electricity_cost(m),
                             "Carbon_emitted": lambda m: self.get_carbon_emitted(m)
                             }

        )

    def override_total_demand(self, total_demand, number_of_agents=None):
        self.total_demand = total_demand
        if total_demand is not None:
            elecsim.scenario.scenario_data.power_plants = self.stratify_data(total_demand)
            demand_modifier = (elecsim.scenario.scenario_data.power_plants.Capacity.sum() /
                               elecsim.scenario.scenario_data.segment_demand_diff[-1]) / 1.6
            logger.info("demand_modifier: {}".format(demand_modifier))
            logger.info(
                "total available capacity: {}".format(elecsim.scenario.scenario_data.power_plants.Capacity.sum()))
            elecsim.scenario.scenario_data.segment_demand_diff = [demand_modifier * demand for demand in
                                                                    elecsim.scenario.scenario_data.segment_demand_diff]

            if number_of_agents is not None:
                total_plants = len(elecsim.scenario.scenario_data.power_plants)
                fraction_to_replace = total_plants/number_of_agents
                company_names = sample(list(elecsim.scenario.scenario_data.power_plants.Company.unique()), number_of_agents)

                company_name_repeated = np.repeat(company_names, int(fraction_to_replace))
                company_name_repeated = np.append(company_name_repeated, np.array(["company_{}".format(number_of_agents-1) for i in range(100)]))

                elecsim.scenario.scenario_data.power_plants.Company = company_name_repeated[:total_plants]

    def override_highest_demand(self, highest_demand):
        if highest_demand:
            elecsim.scenario.scenario_data.initial_max_demand_size = highest_demand

    def override_demand_change(self, demand_change):
        if demand_change:
            elecsim.scenario.scenario_data.yearly_demand_change = demand_change[1:]
            self.demand_change_name = str(demand_change[0]).replace(".", '')
        else:
            self.demand_change_name = "none"

    def override_carbon_scenario(self, carbon_price_scenario):
        if carbon_price_scenario:
            elecsim.scenario.scenario_data.carbon_price_scenario = carbon_price_scenario[1:]
            self.carbon_scenario_name = str(carbon_price_scenario[0]).replace(".", '')
        else:
            self.carbon_scenario_name = "none"

    def write_scenario_data(self):
        if self.step_number == self.max_number_of_steps:
            parent_directory = os.path.dirname(os.getcwd())

            directory = "{}/{}/".format(parent_directory, self.data_folder)
            if not os.path.exists(directory):
                os.makedirs(directory)

            filename = "demand_{}-carbon_{}-datetime_{}-capacity_{}-demand_distribution_{}.csv".format(
                                                                                               self.demand_change_name,
                                                                                               self.carbon_scenario_name,
                                                                                               dt.datetime.now().strftime(
                                                                                                   '%Y-%m-%d_%H-%M-%S'),
                                                                                               elecsim.scenario.scenario_data.segment_demand_diff[
                                                                                                   -1], self.distribution_name)

            directory_filename = "{}/{}.csv".format(directory, filename)

            results_df = self.datacollector.get_model_vars_dataframe()
            results_df.to_csv(directory_filename)

            class TransferData:
                def __init__(self, access_token):
                    self.access_token = access_token

                def upload_file(self, file_from, file_to):
                    """upload a file to Dropbox using API v2
                    """
                    dbx = dropbox.Dropbox(self.access_token)

                    with open(file_from, 'rb') as f:
                        dbx.files_upload(f.read(), file_to)

            access_token = 'J0BrnIaGJ78AAAAAAABLCd6RWS4T1JQwhKCtYcdWTdyE--pvA0-DfNIt4OUnUZQx'
            transferData = TransferData(access_token)

            file_from = "/{}".format(directory_filename)
            file_to = "/{}".format(filename)

            # API v2
            transferData.upload_file(file_from, file_to)






        if self.step_number == self.max_number_of_steps:
            end = perf_counter()
            time_elapased = end - self.start

            self.write_timing_results(end, time_elapased)

    def write_timing_results(self, end, time_elapased):
        if self.time_run:
            timings_data = pd.DataFrame(
                {"time": [time_elapased], "carbon": [elecsim.scenario.scenario_data.carbon_price_scenario[0]],
                 'installed_capacity': [elecsim.scenario.scenario_data.power_plants.Capacity.sum()],
                 'datetime': [dt.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')]})

            parent_directory = os.path.dirname(os.getcwd())

            with open("{}/{}/timing.csv".format(ROOT_DIR, self.data_folder), 'a') as f:
                timings_data.to_csv(f, header=False)
            logger.info("end: {}".format(end))
            logger.info("time_elapsed: {}, carbon: {}, size: {}".format(time_elapased,
                                                                    elecsim.scenario.scenario_data.carbon_price_scenario[
                                                                        0],
                                                                    elecsim.scenario.scenario_data.power_plants.Capacity.sum()))




from elecsim.plants.plant_costs.estimate_costs.estimate_old_plant_cost_params.old_plant_param_calc import OldPlantCosts


class NonFuelOldPlantCosts(OldPlantCosts):

    def __init__(self, year, plant_type, capacity):
        super().__init__(year=year, plant_type=plant_type, capacity=capacity)

    def get_cost_parameters(self):
        """
        Function which estimates the parameters of the non-modern power plant using the LCOE value for the relevant
        year.
        :return: Returns dictionary of the updated parameters for the power plant.
        """
        # List containing parameters to not scale by updated LCOE value. For instance, time taken to build power plant,
        # as they are unrelated to LCOE value.
        self.get_params_to_scale()

        # Multiply values by updated LCOE scale. Conditional based on whether parameter is in params_to_ignore list
        scaled_params = {
            key: (value * self.lcoe_scaler) if key not in self.dict_to_ignore else value
            for key, value in self.estimated_modern_plant_parameters.items()}

        scaled_params.update(self.dict_to_ignore)

        return scaled_params




import pandas as pd
pd.set_option('display.max_columns', 500)
pd.set_option('display.max_rows', 5000)


class UKPlants:

    def __init__(self, uk_plant_db_dir):
        self.plant_db = pd.read_csv(uk_plant_db_dir)

    def modify_plant_db(self):
        """
        Overall method which runs synchronises UK cost and power plant database, removes ambiguous values and removes
        black start motors
        :return: Returns modified UK power plant database
        """

        self._remove_black_start_motors()
        self._plant_type_synchronisation()
        self._remove_ambiguous_values()

        return self.plant_db

    def _plant_type_synchronisation(self):
        """
        Function to match the power plant type plant_type of the UK power plant database with the UK power plant cost database
        :return: Returns modified power plant database
        """
        self.plant_db['Simplified_Type'] = self.plant_db['Fuel'].map(lambda x:
                                                             "Biomass" if "Meat" == x
                                                             else "Offshore" if "Wind (offshore)" == x
                                                             else "Biomass" if "Biomass_poultry_litter" == x
                                                             else "Biomass" if "Straw" == x
                                                             else "Recip_diesel" if "Diesel" == x
                                                             else "Biomass" if "Biomass_wood" == x
                                                             else "Recip_gas" if "Gas" == x
                                                             else "PV" if "Solar" == x
                                                             else "OCGT" if "Gas oil" == x
                                                             else "EfW" if "Waste" == x
                                                             else "Hydro_Store" if "Pumped storage" == x
                                                             else "Onshore" if "Wind" == x
                                                             else x)

    def _remove_ambiguous_values(self):
        """
        Function which removes ambiguous values from the database.
        :return: Returns modified power plant database
        """
        # Dolgarrog start date labelled as 1926/2002, replaced this with 2002
        self.plant_db.loc[418, 'Start_date'] = '2002.00'

        # Killingholme power station was demolished and closed in 2016.
        self.plant_db = self.plant_db[self.plant_db.Name != 'Killingholme']


    def _remove_black_start_motors(self):
        """
        Removal of power plants due to the fact that they do not contribute to national grid electricity supply,
        and are used as starter motors in the event of an electric grid blackout.
        :return: Returns modified power plant database
        """
        # self.plant_db = self.plant_db[self.plant_db.Name != 'Baglan Bay OCGT']
        self.plant_db = self.plant_db[~self.plant_db['Name'].isin(['Baglan Bay OCGT','Drax GT','West Burton GT',
                                                                   'Coolkeeragh OCGT','Aberthaw GT','Didcot GT',
                                                                   'Little Barford GT', 'Fiddler’s Ferry GT',
                                                                   'Keadby GT', "Grain GT", 'Ratcliffe GT'])]


pp_dat = UKPlants("/Users/b1017579/Documents/PhD/Projects/10. ELECSIM/10. ELECSIM/data/Power_Plants/No_Location/power_plants_removed_location.csv")
# plant_db = pp_dat.remove_black_start_motors()

plant_db = pp_dat.modify_plant_db()
plant_db.to_csv('/Users/b1017579/Documents/PhD/Projects/10. ELECSIM/10. ELECSIM/data/Power_Plants/No_Location/power_plant_db_with_simplified_type.csv')



"""
File name: __init__.py
Date created: 08/11/2019
Feature: #Enter feature description here
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"



import mysql.connector
from mysql.connector import errorcode

import os.path
import sys
import ray


sys.path.append(os.path.join(os.path.dirname(__file__), '../../..'))

from elecsim.model.world import World
import tracemalloc

import pandas as pd
import linecache

from elecsim.constants import ROOT_DIR
import string
from deap import algorithms
from deap import base
from deap import benchmarks
from deap.benchmarks.tools import diversity, convergence, hypervolume
from deap import creator
from deap import tools

import array
import random
import logging
logger = logging.getLogger(__name__)
import numpy as np
import numpy

from scoop import futures

from pebble import ProcessPool
from concurrent.futures import TimeoutError

import time
from pathlib import Path
project_dir = Path("__file__").resolve().parents[1]
import string
import random


import array
import random
import json

import numpy

from math import sqrt

from deap import algorithms
from deap import base
from deap import benchmarks
from deap.benchmarks.tools import diversity, convergence, hypervolume
from deap import creator
from deap import tools

creator.create("FitnessMin", base.Fitness, weights=(-1.0, -1.0))
creator.create("Individual", array.array, typecode='d', fitness=creator.FitnessMin)

@ray.remote(num_return_vals=2)
def world_eval(individual, carbon_policy):
    beis_params = [0.00121256259168, 46.850377392563864, 0.0029982421515, 28.9229765616468, 0.00106156336814, 18.370337670063762, 0.00228312539654, 0.0, 0.0024046471141100003, 34.43480109190594, 0.0, -20.88014916953091, 0.0, 8.15032953348701, 0.00200271495761, -12.546185375581802, 0.00155518243668, 39.791132970522796, 0.00027449937576, 8.42878689508516, 0.00111989525697, 19.81640207212787, 0.00224091998324, 5.26288570922149, 0.00209189353332, -5.9117317131295195, 0.00240696026847, -5.0144941135222, 0.00021183142492999999, -1.29658413335784, 0.00039441444392000004, -11.41659250225168, 0.00039441444392000004, -11.41659250225168, 120.21276910611674, 0.0, 0.00059945111227]
    # beis_params = [0.00121256259168, 46.850377392563864, 0.0029982421515, 28.9229765616468, 0.00106156336814, 18.370337670063762, 0.00228312539654, 0.0, 0.0024046471141100003, 34.43480109190594, 0.0, -20.88014916953091, 0.0, 8.15032953348701, 0.00200271495761, -12.546185375581802, 0.00155518243668, 39.791132970522796, 0.00027449937576, 8.42878689508516, 0.00111989525697, 19.81640207212787, 0.00224091998324, 5.26288570922149, 0.00209189353332, -5.9117317131295195, 0.00240696026847, -5.0144941135222, 0.00021183142492999999, -1.29658413335784, 0.00039441444392000004, -11.41659250225168, 0.0021988838824299997, 12.633572943294599, 120.21276910611674, 0.0, 0.00059945111227]

    prices_individual = np.array(beis_params[:-3]).reshape(-1, 2).tolist()

    MARKET_TIME_SPLICES = 8
    YEARS_TO_RUN = 17
    number_of_steps = YEARS_TO_RUN * MARKET_TIME_SPLICES

    scenario_2018 = "{}/../run/beis_case_study/scenario/reference_scenario_2018.py".format(ROOT_DIR)

    # if individual[0] == 1:
    #     individual = [individual[1]*i + individual[2] for i in range(1, 20)]
    # else:
    #     individual = [individual[3]*i ** individual[4] + individual[2] for i in range(1, 20)]
    individual = [individual[0]*i + individual[1] for i in range(1, 20)]
    print(individual)

    world = World(carbon_price_scenario=individual[:-1], initialization_year=2018, scenario_file=scenario_2018, market_time_splices=MARKET_TIME_SPLICES, data_folder="{}_data".format(carbon_policy), number_of_steps=number_of_steps, long_term_fitting_params=prices_individual, highest_demand=63910, nuclear_subsidy=beis_params[-3], future_price_uncertainty_m=beis_params[-2], future_price_uncertainty_c=beis_params[-1])
    for _ in range(YEARS_TO_RUN):
        for i in range(MARKET_TIME_SPLICES):
            # try:
            if i/8 == 0:
                print('end of year')

            average_electricity_price, carbon_emitted = world.step()
            # except Exception as e:
            #     print("try catch error: \n{}".format(e))
            #     return 889999999999999999, 889999999999999999
            # if carbon_emitted is bool:
            if isinstance(carbon_emitted, bool):
                print("carbon emitted error")
                return 999999999999999999, 999999999999999999

            if not isinstance(average_electricity_price, float):
                print("average_electricity_price {}, type {}".format(average_electricity_price, type(average_electricity_price)))
                average_electricity_price = 779999999999999999
            if not isinstance(carbon_emitted, float):
                print("carbon_emitted {}, type {}".format(carbon_emitted, type(carbon_emitted)))
                carbon_emitted = 779999999999999999

    print("average_electricity_price: {}, carbon_emitted: {}".format(average_electricity_price, carbon_emitted))
    return average_electricity_price, carbon_emitted
toolbox = base.Toolbox()

# Problem definition
# Functions zdt1, zdt2, zdt3, zdt6 have bounds [0, 1]
BOUND_LOW, BOUND_UP = 0, 250

# Functions zdt4 has bounds x1 = [0, 1], xn = [-5, 5], with n = 2, ..., 10
# BOUND_LOW, BOUND_UP = [0.0] + [-5.0]*9, [1.0] + [5.0]*9

# Functions zdt1, zdt2, zdt3 have 30 dimensions, zdt4 and zdt6 have 10
NDIM = 19

def uniform(low, up, size=None):
    try:
        return [random.uniform(a, b) for a, b in zip(low, up)]
    except TypeError:
        return [random.uniform(a, b) for a, b in zip([low] * size, [up] * size)]
#



# toolbox.register("attr_float", uniform, BOUND_LOW, BOUND_UP, NDIM)

# toolbox.register("attr_function", random.randint, 0, 1)
toolbox.register("attr_m", random.uniform, -14, 14)
toolbox.register("attr_c", random.uniform, 0, 250)

# toolbox.register("attr_L", random.uniform, -10, 10)
# toolbox.register("attr_k", random.uniform, 1, 14)
# toolbox.register("attr_n", random.uniform, 0, 250)
# toolbox.register("individual", tools.initCycle, creator.Individual, (toolbox.attr_function, toolbox.attr_m, toolbox.attr_c, toolbox.attr_a, toolbox.attr_d), n=1)
toolbox.register("individual", tools.initCycle, creator.Individual, (toolbox.attr_m, toolbox.attr_c), n=1)
toolbox.register("population", tools.initRepeat, list, toolbox.individual)

toolbox.register("evaluate", world_eval)
# toolbox.register("mate", tools.cxSimulatedBinaryBounded, low=BOUND_LOW, up=BOUND_UP, eta=20.0)
toolbox.register("mate", tools.cxTwoPoint)
toolbox.register("mutate", tools.mutFlipBit, indpb=0.05)
toolbox.register("select", tools.selNSGA2)
toolbox.register("map_distributed", futures.map)


config = {
  'host':'elecsimresults2.mysql.database.azure.com',
  'user':'alexkell@elecsimresults2',
  'password':'b3rz0s4m4dr1dth3h01113s!',
  'database':'carbonoptimiser',
  'ssl_ca':'run/validation-optimisation/database/BaltimoreCyberTrustRoot.crt.pem'
  # 'ssl_ca':'/Users/b1017579/Documents/PhD/Projects/10-ELECSIM/run/validation-optimisation/database/BaltimoreCyberTrustRoot.crt.pem'

}


def main(seed=None):
    random.seed(seed)

    NGEN = 999
    # MU = 120
    MU = 80
    # MU = 4
    CXPB = 0.9

    stats = tools.Statistics(lambda ind: ind.fitness.values)
    # stats.register("avg", numpy.mean, axis=0)
    # stats.register("std", numpy.std, axis=0)
    stats.register("min", numpy.min, axis=0)
    stats.register("max", numpy.max, axis=0)

    logbook = tools.Logbook()
    logbook.header = "gen", "evals", "std", "min", "avg", "max"

    pop = toolbox.population(n=MU)

    # Evaluate the individuals with an invalid fitness
    invalid_ind = [ind for ind in pop if not ind.fitness.valid]
    fitnesses = toolbox.map_distributed(toolbox.evaluate, invalid_ind)
    for ind, fit in zip(invalid_ind, fitnesses):
        ind.fitness.values = fit

    # This is just to assign the crowding distance to the individuals
    # no actual selection is done
    pop = toolbox.select(pop, len(pop))

    record = stats.compile(pop)
    logbook.record(gen=0, evals=len(invalid_ind), **record)
    print(logbook.stream)

    # Begin the generational process
    for gen in range(1, NGEN):
        # Vary the population
        offspring = tools.selTournamentDCD(pop, len(pop))
        offspring = [toolbox.clone(ind) for ind in offspring]

        for ind1, ind2 in zip(offspring[::2], offspring[1::2]):
            if random.random() <= CXPB:
                toolbox.mate(ind1, ind2)

            toolbox.mutate(ind1)
            toolbox.mutate(ind2)
            del ind1.fitness.values, ind2.fitness.values

        # Recalculate all individual due to stochastic nature of simulation
        for ind in offspring:
            del ind.fitness.values

        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        fitnesses = toolbox.map_distributed(toolbox.evaluate, invalid_ind)
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

        # Select the next generation population
        pop = toolbox.select(pop + offspring, MU)
        record = stats.compile(pop)
        logbook.record(gen=gen, evals=len(invalid_ind), **record)
        print(logbook.stream)

        front = numpy.array(
            [ind.fitness.values + tuple(ind) for ind in pop])

        print("front: {}".format(front))

        try:
            conn = mysql.connector.connect(**config)
            print("Connection established")
        except mysql.connector.Error as err:
            if err.errno == errorcode.ER_ACCESS_DENIED_ERROR:
                print("Something is wrong with the user name or password")
            elif err.errno == errorcode.ER_BAD_DB_ERROR:
                print("Database does not exist")
            else:
                print(err)
        else:
            cursor = conn.cursor()

        # first_part = 'INSERT INTO carbon_results (reward,carbon_1,carbon_2,carbon_3,carbon_4,carbon_5,carbon_6,carbon_7,carbon_8,carbon_9,carbon_10,carbon_11,carbon_12,carbon_13,carbon_14,carbon_15,carbon_16,carbon_17,carbon_18) VALUES '
        #
        # insert_vars = "".join(["({},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{}),\n".format(ind.flat[0], ind.flat[1], ind.flat[2], ind.flat[3], ind.flat[4], ind.flat[5], ind.flat[6], ind.flat[7], ind.flat[8], ind.flat[9], ind.flat[10], ind.flat[11], ind.flat[12], ind.flat[13], ind.flat[14], ind.flat[15], ind.flat[16], ind.flat[17], ind.flat[18]) for ind in front])

        # first_part = 'INSERT INTO carbon_results_function (average_electricity_price,carbon_emitted,attr_function,attr_m,attr_c,attr_a,attr_d) VALUES '
        #
        # insert_vars = "".join(["({},{},{},{},{},{},{}),\n".format(ind.flat[0], ind.flat[1], ind.flat[2], ind.flat[3], ind.flat[4], ind.flat[5], ind.flat[6]) for ind in front])
        first_part = 'INSERT INTO carbon_results_function (average_electricity_price,carbon_emitted,attr_m,attr_c) VALUES '

        insert_vars = "".join(["({},{},{},{}),\n".format(ind.flat[0], ind.flat[1], ind.flat[2], ind.flat[3]) for ind in front])



        insert_cmd = first_part+insert_vars
        insert_cmd = insert_cmd[:-2]
        # print("command: {}".format(insert_cmd))

        cursor.execute(insert_cmd)
        conn.commit()
        cursor.close()
        conn.close()


    print("Final population hypervolume is %f" % hypervolume(pop, [11.0, 11.0]))

    return pop, logbook

if __name__ == "__main__":
    # with open("pareto_front/zdt1_front.json") as optimal_front_data:
    #     optimal_front = json.load(optimal_front_data)
    # Use 500 of the 1000 points in the json file
    # optimal_front = sorted(optimal_front[i] for i in range(0, len(optimal_front), 2))

    # pop, stats = main()

    # #lowest
    # world_eval([11.384098, 81.34436])
    #
    # #highest
    # world_eval([11.384098, 202.53813])
    #
    # #flat
    # world_eval([-0.2289701, 192.5701])

    def world_eval_parallel(params, strategy):
        output1, output2 = [], []

        # Launch the tasks.
        for j in range(100):
            id1, id2 = world_eval.remote(params, strategy)
            output1.append(id1)
            output2.append(id2)

        # Block until the results have finished and get the results.
        output1 = ray.get(output1)
        output2 = ray.get(output2)

    ray.init(num_cpus=os.cpu_count()-1)
    # world_eval_parallel([11.384098, 81.34436], "lowest")
    # world_eval_parallel([11.384098, 202.53813], "highest")
    world_eval_parallel([-0.2289701, 192.5701], "flat")




from functools import lru_cache

from numpy import isnan
from numpy import ndarray
import numpy as np
from sklearn.linear_model import LinearRegression

import logging

import elecsim.scenario.scenario_data
from elecsim.plants.plant_costs.estimate_costs.estimate_modern_power_plant_costs.predict_modern_plant_costs import PredictModernPlantParameters
from elecsim.plants.plant_costs.estimate_costs.estimate_old_plant_cost_params.fuel_plant_calculations.fuel_plants_old_params import FuelOldPlantCosts
from elecsim.plants.plant_costs.estimate_costs.estimate_old_plant_cost_params.non_fuel_plant_calculations.non_fuel_plants_old_params import NonFuelOldPlantCosts
from elecsim.plants.plant_registry import PlantRegistry

logger = logging.getLogger(__name__)
"""
File name: _select_cost_estimator
Date created: 01/12/2018
Feature: # Functionality to estimate costs based on year. If year is past 2018 then use modern data from BEIS file.
         # If data is historic, then predict from historic LCOE values, maintaining same ratios from 2018.
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"

@lru_cache(maxsize=10000)
def create_power_plant(name, start_date, simplified_type, capacity):
    """
    Functionality that estimates the cost of a power plant based solely on year of construction, type of plant and capacity.
    :param name: Name of power plant.
    :param start_date: Starting year of power plant
    :param simplified_type: Type of power plant
    :param capacity: Capacity of power plant
    :return: Power plant object initialized with relevant variables.
    """
    estimated_cost_parameters = _select_cost_estimator(start_year=start_date,
                                                       plant_type=simplified_type,
                                                       capacity=capacity)
    power_plant_obj = PlantRegistry(simplified_type).plant_type_to_plant_object()
    power_plant = power_plant_obj(name=name, plant_type=simplified_type,
                                  capacity_mw=capacity, construction_year=start_date,
                                  **estimated_cost_parameters)
    return power_plant

# def create_power_plant_group(name, money, start_date, simplified_type):
def create_power_plant_group(name, start_date, simplified_type, capacity, number_of_plants_to_purchase):
    # capacity = predict_capacity_from_money(money, simplified_type, capacity, start_date)
    plant_object = create_power_plant(name, start_date=start_date, simplified_type=simplified_type, capacity=capacity*number_of_plants_to_purchase)

    # plant_object = create_power_plant(name, start_date=start_date, simplified_type=simplified_type, capacity=capacity)
    return plant_object




# def predict_capacity_from_money(money, simplified_type, capacity, start_date):
#     test_capacity = [capacity, capacity+0.5]
#     plant_1 = create_power_plant("invested_plant", start_date, simplified_type, test_capacity[0])
#     plant_2 = create_power_plant("invested_plant", start_date, simplified_type, test_capacity[1])
#     y = np.array([plant_1.get_upfront_costs(), plant_2.get_upfront_costs()]).reshape(-1, 1)
#     # test_capacity_array = np.array(test_capacity).reshape(-1, 1)
#     test_capacity_array = test_capacity
#     reg = LinearRegression().fit(y, test_capacity_array)
#     size = reg.predict(np.array(money).reshape(-1, 1))
#     return int(size)


def _select_cost_estimator(start_year, plant_type, capacity):
    earliest_modern_plant_year = min([int(column.split(" _")[1]) for column in elecsim.scenario.scenario_data.modern_plant_costs.filter(regex='Connect_system_cost-Medium _').columns])
    _check_digit(capacity, "capacity")
    _check_digit(start_year, "start_year")
    _check_positive(start_year, "start_year")
    _check_positive(capacity, "capacity")

    hist_costs = elecsim.scenario.scenario_data.power_plant_historical_costs_long
    hist_costs = hist_costs[hist_costs.Technology.map(lambda x: x in plant_type)].dropna()
    if start_year < earliest_modern_plant_year and not hist_costs.empty:
        require_fuel = PlantRegistry(plant_type).check_if_fuel_required()
        cost_parameters = _estimate_old_plant_cost_parameters(capacity, plant_type, require_fuel, start_year)
        _check_parameters(capacity, cost_parameters, plant_type, start_year)
        return cost_parameters
    else:
        cost_parameters = PredictModernPlantParameters(plant_type, capacity, start_year).parameter_estimation()
        _check_parameters(capacity, cost_parameters, plant_type, start_year)
        return cost_parameters


def _check_parameters(capacity, cost_parameters, plant_type, start_year):
    assert not all(value == 0 for value in
                   cost_parameters.values()), "All values are 0 for cost parameters for power plant of year {}, type {}, and capacity {}".format(
        start_year, plant_type, capacity)
    assert not any(isnan(value).any() for value in
                   cost_parameters.values()), "All values are nan for cost parameters for power plant of year {}, type {}, and capacity {}".format(
        start_year, plant_type, capacity)


def _check_digit(value, string):
    if not isinstance(value, int) and not isinstance(value, float) and not isinstance(value, ndarray):
        raise ValueError("{} must be a number".format(string))

def _check_positive(variable, string):
    if variable < 0:
        raise ValueError("{} must be greater than 0. Produced is: {}".format(string, variable))


def _estimate_old_plant_cost_parameters(capacity, plant_type, require_fuel, start_year):
    if require_fuel:
        fuel_plant_parameters = FuelOldPlantCosts(start_year, plant_type, capacity)
        return fuel_plant_parameters.estimate_cost_parameters()
    else:
        non_fuel_plant_parameters = NonFuelOldPlantCosts(start_year, plant_type, capacity)
        return non_fuel_plant_parameters.get_cost_parameters()



import logging
from itertools import chain
from random import shuffle

from mesa.time import BaseScheduler

logging.getLogger(__name__)

from elecsim.agents.generation_company.gen_co import GenCo
from elecsim.agents.demand.demand import Demand
from elecsim.market.electricity.market.power_exchange import PowerExchange
import elecsim.scenario.scenario_data
from elecsim.role.market.latest_market_data import LatestMarketData

class OrderedActivation(BaseScheduler):
    """ A scheduler which activates each agent in the order that they are added to the scheduler

    Assumes that all agents have a step(model) method.

    """
    def step(self):
        """ Executes the step of all agents, one at a time.
        """

        gen_cos = [agent for agent in self.agents if isinstance(agent, GenCo)]
        shuffle(gen_cos)

        demand_agents = [agent for agent in self.agents if isinstance(agent, Demand)]

        if elecsim.scenario.scenario_data.investment_mechanism == "RL" and self.model.step_number % self.model.market_time_splices == 0:
            obs = LatestMarketData(self.model).get_RL_investment_observations()
            actions = self.model.client.get_action(self.model.eid, obs)
            # logger.info("actions: {}".format(actions))

            reward = {}
            for gen_co_name, action in actions.items():
                gen_co_to_invest = [gen_co for gen_co in gen_cos if gen_co.name == gen_co_name][0]

                gen_co_to_invest.invest_RL(action)

                reward[gen_co_name] = gen_co_to_invest.money

        for agent in chain(gen_cos, demand_agents):
            agent.step()

        reward = {}
        for genco in self.model.get_gencos():
            reward[genco.name] = genco.money

        if elecsim.scenario.scenario_data.investment_mechanism == "RL":
            self.model.client.log_returns(self.model.eid, reward=reward)

        self.steps += 1
        self.time += 1




from setuptools import setup

"""
File name: setup
Date created: 19/02/2019
Feature: #Enter feature description here
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"

setup(name='gym_elecsim',
      version='0.0.1',
      install_requires=['gym']  # And any other dependencies foo needs
)



import logging

from elecsim import constants as constants

logger = logging.getLogger(__name__)
"""
File name: calculate_lcoe
Date created: 18/12/2018
Feature: #Enter feature description here
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"


class PlantCostCalculations:
    def __init__(self, capacity_mw, construction_year, average_load_factor, pre_dev_period, construction_period, operating_period, pre_dev_spend_years, construction_spend_years, pre_dev_cost_per_mw, construction_cost_per_mw, infrastructure, fixed_o_and_m_per_mw, variable_o_and_m_per_mwh, insurance_cost_per_mw, connection_cost_per_mw):
        self.capacity_mw = capacity_mw

        self.construction_year = construction_year

        self.average_load_factor = average_load_factor

        self.pre_dev_period = pre_dev_period
        self.pre_dev_spend_years = pre_dev_spend_years

        self.construction_period = construction_period
        self.construction_spend_years = construction_spend_years

        self.pre_dev_cost_per_mw = pre_dev_cost_per_mw

        self.construction_cost_per_mw = construction_cost_per_mw

        self.operating_period = operating_period

        self.infrastructure = infrastructure

        self.fixed_o_and_m_per_mw = fixed_o_and_m_per_mw

        self.variable_o_and_m_per_mwh = variable_o_and_m_per_mwh

        self.insurance_cost_per_mw = insurance_cost_per_mw

        self.connection_cost_per_mw = connection_cost_per_mw

    @property
    def infrastructure(self):
        return self._infrastructure

    @infrastructure.setter
    def infrastructure(self, value):
        self._infrastructure = value * constants.KW_TO_MW

    def _discount_data(self, variable, discount_rate):
        discount_multiplier = [1]

        for _ in range(len(variable)):
            discount_multiplier.extend([discount_multiplier[-1]/(1+discount_rate)])

        discounted_var = [a*b for a,b in zip(variable, discount_multiplier)]
        return discounted_var

    def _pre_dev_yearly_spend(self):
        """
        Calculate the yearly pre-development spend
        :return: List containing the spend per year for pre-development costs.
        """
        pre_dev_cost_total = self.pre_dev_cost_per_mw * self.capacity_mw # Total construction costs for power plant

        pre_dev_spend_per_year = [x*pre_dev_cost_total for x in self.pre_dev_spend_years]  # Creates a list containing pre development spend per year

        return pre_dev_spend_per_year

    def _construction_yearly_spend(self):
        """

        Calculate the cost of yearly construction spend. Includes infrastructure cost during the final year of construction.
        :return: List containing spend per year for construction
        """

        construction_cost_total = self.construction_cost_per_mw * self.capacity_mw # Total construction costs for power plant
        construction_spend_per_year = [x*construction_cost_total for x in self.construction_spend_years]  # Creates a list containing construction spend per year

        return construction_spend_per_year
        # infrastructure_gbp = self.infrastructure
        # construction_spend_per_year[-1] += infrastructure_gbp  # Infrastructure cost is taken into account in the final year of construction

        # return construction_spend_per_year

    def _capex(self):
        """
        Calculation of capital expenditure, which includes insurance, building and pre-development costs. This is done
        by adding pre-development yearly cost with construction yearly spend.
        :return: Capital expenditure cost
        """
        capex = self._pre_dev_yearly_spend() + self._construction_yearly_spend()
        capex[-1]+=self.infrastructure
        return capex

    def _insurance_cost(self):
        """
        Calculates the yearly insurance cost
        :return: List containing insurance cost for each year of the operating period
        """
        insurance_cost_total = [0] * int(self.pre_dev_period+self.construction_period) +[self.insurance_cost_per_mw * self.capacity_mw] * int(self.operating_period) # Calculation of insurance cost for this instance of a power plant
        return insurance_cost_total

    def _variable_o_and_m_cost(self):
        """
        Calculates the operating and maintenance cost per year. Makes an assumption based on average load factor.
        :return: List containing variable operating and maintenance cost for each year of the operating period.
        """
        variable_o_and_m_cost_per_year = self.variable_o_and_m_per_mwh * self.capacity_mw * 365 * 24 * self.average_load_factor  # Variable cost is calculated as a function of time plant is utilised in a year
        variable_o_and_m_cost_per_year = [0] * int(self.pre_dev_period+self.construction_period) + [variable_o_and_m_cost_per_year] * int(self.operating_period)
        return variable_o_and_m_cost_per_year

    def _fixed_o_and_m_cost(self):
        """
        Calculates the fixed operating and maintenance cost per year.
        :return: List containing fixed operating and maintenance cost for each year of the operating period.
        """
        fixed_o_and_m_per_mw_cost_total = self.fixed_o_and_m_per_mw * self.capacity_mw  # Fixed cost calculated for this instance of power plant
        fixed_o_and_m_per_mw_cost_total = [0] * int(self.pre_dev_period+self.construction_period) + [fixed_o_and_m_per_mw_cost_total] * int(self.operating_period)
        return fixed_o_and_m_per_mw_cost_total

    def _opex_cost(self):
        """
        Calculation of operating expenditure, which includes fixed and variable costs
        :return: Operating expenditure cost
        """

        insur_cost = self._insurance_cost()
        var_o_and_m_cost = self._variable_o_and_m_cost()
        fixed_o_and_m_cost = self._fixed_o_and_m_cost()

        opex = [sum(x) for x in zip(insur_cost, var_o_and_m_cost, fixed_o_and_m_cost)]

        return opex

    def _electricity_generated(self):
        """
        Estimates the amount of electricity generated over the lifetime of the project based on the average load factor
        :return: Returns a list containing the electricity generated per year
        """
        HOURS_IN_DAY = constants.HOURS_IN_DAY
        DAYS_IN_YEAR = constants.DAYS_IN_YEAR

        elec_gen = [self.capacity_mw * self.average_load_factor * HOURS_IN_DAY * DAYS_IN_YEAR] * int(self.operating_period)
        elec_gen = [0] * int(self.pre_dev_period+self.construction_period) + elec_gen

        return elec_gen

    def total_income(self, expected_sale_price):
        beginning_year_operation = self.construction_year
        end_of_lifetime_year = int(beginning_year_operation)+int(self.operating_period)+int(self.pre_dev_period+self.construction_period)
        years_of_plant_operation = range(int(beginning_year_operation), end_of_lifetime_year)

        yearly_return = self.capacity_mw*self.average_load_factor*365*24*expected_sale_price
        returns = [yearly_return]*(len(years_of_plant_operation)-int(self.construction_period+self.pre_dev_period))
        years_not_running = [0]*int(self.construction_period+self.pre_dev_period)
        expected_returns = years_not_running + returns
        return expected_returns

    def calculate_yearly_outflow(self):

        pre_development_cost = self.pre_dev_cost_per_mw * self.pre_dev_spend_years * self.capacity_mw
        construction_cost = self.pre_dev

        capital_cost = self.capacity_mw * (self.pre_dev_cost_per_mw + self.construction_cost_per_mw) + self.infrastructure
        logger.debug("capital cost: {}".format(capital_cost))
        yearly_capital_cost = capital_cost / self.operating_period
        logger.debug("yearly capital cost: {}".format(yearly_capital_cost))

        yearly_fixed_costs = self.capacity_mw* (self.fixed_o_and_m_per_mw + self.insurance_cost_per_mw + self.connection_cost_per_mw)

        logger.debug("yearly fixed costs: {}".format(yearly_fixed_costs))

        negative_yearly_cash_flow = yearly_fixed_costs + yearly_capital_cost
        return negative_yearly_cash_flow



import logging
from math import log

logger = logging.getLogger(__name__)

"""
File name: renewable_learning_rate
Date created: 21/12/2018
Feature: # Contains the functionality for implementing a learning rate for renewable plants to simulate a decrease in prices
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"

logging.basicConfig(level=logging.DEBUG)

def future_renewable_energy_costs(starting_lcoe, learning_rate, number_of_generation_assets):
    next_price = starting_lcoe*number_of_generation_assets**(log(learning_rate, 2))
    return next_price



from scipy.stats import linregress
import numpy as np
from functools import lru_cache
import logging

logger = logging.getLogger(__name__)
"""
File name: linear_regression
Date created: 29/12/2018
Feature: #Enter feature description here
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"


@lru_cache(1024)
def linear_regression(regression, years_to_look_back, years_to_look_ahead=None):
    regression = list(regression)
    years_to_regress = list(range(years_to_look_back))
    years_to_regress = np.array(years_to_regress).astype(float)
    try:
        regression = np.stack(regression, axis=1)
    except:
        pass
    try:
        m, c, _, _, _ = linregress(years_to_regress, np.array(regression).astype(float))
        next_value = m * (years_to_look_back + years_to_look_ahead-1) + c
    except Exception as e:
        logger.warning(str(e.with_traceback))
        logger.warning("regression: {}".format(regression))
        logger.warning("years to regress: {}".format(years_to_regress))
        next_value = regression[-1]
    return next_value



import logging
import pandas as pd
import elecsim.scenario.scenario_data
from elecsim.agents.demand.demand import Demand
logger = logging.getLogger(__name__)



"""demand.py: Agent which simulates the demand of the UK"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "Alexander@Kell.es"


class MultiDayDemand(Demand):
    """ An agent representing UK electricity demand by selecting representative days.
    """

    def __init__(self, model, unique_id, multi_year_data):
        super().__init__(model=model, unique_id=unique_id)

        self.yearly_demand_change = elecsim.scenario.scenario_data.yearly_demand_change
        self.demand = multi_year_data[multi_year_data.data_type == "load"]
        self.demand_ldc = self.demand.groupby('cluster').apply(lambda x: x.sort_values('capacity_factor', ascending=False).reset_index())
        self.demand_ldc.capacity_factor = self.demand_ldc.capacity_factor.values*elecsim.scenario.scenario_data.initial_max_demand_size

        self.segment_hours = 0
        self.segment_consumption = 0

        self.years_from_start = 0
        self.steps_from_start = 0
        self.year_segment_hours = 0
        self.year_segment_consumption = 0


    def step(self):
        logger.debug("Stepping demand")
        steps_since_year = self.model.step_number % self.model.market_time_splices

        if steps_since_year == 0:
            self.demand_ldc.capacity_factor = self.demand_ldc.capacity_factor * self.yearly_demand_change[self.years_from_start]
            self.year_segment_hours, self.year_segment_consumption = self.set_demand_for_year()

        grouped_days = self.demand_ldc.reset_index(drop=True).groupby("cluster")
        representative_day = grouped_days.get_group((list(grouped_days.groups)[steps_since_year]))

        self.segment_consumption = representative_day.capacity_factor.tolist()
        self.segment_hours = representative_day.cluster.cumsum().tolist()

        # self.steps_from_start += 1

        # load to change each year due to certain scenario

    def set_demand_for_year(self):

        # grouped_days = self.demand_ldc.reset_index(drop=True).groupby("cluster")

        demand_dataframe = self.demand_ldc.sort_values("capacity_factor", ascending=False)

        demand_dataframe['hour'] = demand_dataframe.cluster.cumsum()



        # year_segment_consumption = demand_dataframe.capacity_factor.tolist()
        # year_segment_hours = demand_dataframe.hour.tolist()

        year_segment_consumption = (pd.DataFrame(pd.cut(demand_dataframe.capacity_factor, 20*self.model.market_time_splices).apply(lambda x: x.right)).groupby("capacity_factor").first()).index.tolist()
        year_segment_hours = (pd.DataFrame(pd.cut(demand_dataframe.hour, 20*self.model.market_time_splices).apply(lambda x: x.right)).groupby("hour").first()).index.tolist()

        # final_day = 0
        # for _, day in grouped_days:
        #     year_segment_consumption.append(day.capacity_factor.tolist())
        #     segment_hours = [final_day + hour * day.counts.iloc[0] for hour in range(1, 25)]
        #     year_segment_hours.append(segment_hours)
        #     final_day = 24 * day.counts.iloc[0]

        # year_segment_consumption = demand_dataframe.

        return year_segment_hours, year_segment_consumption




import logging
logger = logging.getLogger(__name__)

"""Bid.py: A class which holds information for each bid"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "Alexander@Kell.es"


class Bid:

    def __init__(self, gen_co, plant, segment_hours, capacity_bid, price_per_mwh, year_of_bid):
        """Bid class that contains all the information related to the creation of bids

        :param gen_co: Generation company from which the bid originates from
        :param plant: Power plant from which the bid originates from
        :param segment_hours: Number of hours sent for bid
        :param capacity_bid: Electricity consumption bid amount
        :param price_per_mwh: Price bid per Megawatt Hour
        """
        self.gen_co = gen_co
        self.plant = plant
        self.segment_hours = segment_hours
        self.capacity_bid = capacity_bid
        self.price_per_mwh = price_per_mwh
        self.year_of_bid = year_of_bid
        self.price_bid = price_per_mwh

        self.bid_accepted = False
        self.bid_rejected = False
        self.partly_accepted = False

    def reject_bid(self, segment_hour):
        """
        Function to be called when bid is rejected
        :return: None
        """
        self.plant.capacity_fulfilled[segment_hour] = 0
        self.bid_rejected = True

    def accept_bid(self, segment_hour):
        """
        Function to be called when bid is accepted
        :return: None
        """
        # Update capacity of plant once bid is accepted

        # segment_hour = str(segment_hour)
        self.plant.capacity_fulfilled[segment_hour] = self.capacity_bid

        self.bid_accepted = True
        self.plant.accepted_bids.append(self)

        if self.plant.plant_type == "Nuclear" and self.gen_co.model.nuclear_subsidy is not None:
            self.price_per_mwh += self.gen_co.model.nuclear_subsidy

    def partially_accept_bid(self, segment_hour, demand_fulfilled):
        """
        Function to be called when bid is partially accepted
        :param demand_fulfilled:
        :return: None
        """
        # Update capacity of plant once bid is partly accepted
        self.plant.capacity_fulfilled[segment_hour] = demand_fulfilled
        self.partly_accepted = True
        self.plant.accepted_bids.append(self)



    def __str__(self):
        return "Plant type: " + self.plant.plant_type + ", Min running time: " +str(self.plant.min_running)+", Number of hours: "+str(self.segment_hours)+", Capacity Bid: "+str(self.capacity_bid)+", Price per MW: "+str(self.price_per_mwh) + ", Price Bid: "+str(self.price_bid)+", Plant: " + self.plant.__repr__()

    def to_dict(self):
        return {
            'plant_type': self.plant.plant_type,
            'segment_hours': self.segment_hours,
            'capacity_bid': self.capacity_bid
        }



from distutils.core import setup
from setuptools import find_packages

import os

def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

extra_files = package_files('elecsim/data/processed/')
print(extra_files)

setup(
  name = 'elecsim',
  packages = find_packages(exclude=['tests*', 'tests.*']),
  package_data={'elecsim': extra_files},

  py_modules=['elecsim'],
  version = '0.1.78',
  license='MIT',
  description = 'Agent-based Model for Electricity Markets',
  author = 'Alexander Kell',
  author_email = 'alexander@kell.es',
  url = 'https://github.com/alexanderkell/elecsim',
  download_url = 'https://github.com/alexanderkell/elecsim/archive/0.1.1.1.tar.gz',
  keywords = ['ElecSIM', 'Agent-based', 'Electricity', 'Market', 'Model', 'Simulation'],
    install_requires=[
        'mesa',
        'pandas',
        'scipy',
        'pathos'
    ],
  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
  ],
)



import os.path
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '../..'))

from mesa.batchrunner import BatchRunnerMP

from elecsim.model.world import World


from elecsim.scenario.scenario_data import power_plants
import logging

logger = logging.getLogger(__name__)

logging.basicConfig(level=logging.INFO)

"""
File name: batch_run_timer
Date created: 25/01/2019
Feature: # Functionality to time run of model at different sized countries
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"


class DemandTimer:

    def __init__(self):
        self.power_plants = power_plants


    def run_world_with_demand_and_power_plants(self):

        number_of_steps = 40
        data_folder = "minimum_sized_country_3"

        fixed_params = {
            "initialization_year": 2018,
            "number_of_steps": number_of_steps,
            "demand_change": [1.0] * 50,
            "carbon_price_scenario": [10]*50,
            "data_folder": data_folder,
            "time_run":False
        }

        variable_params = {
            "total_demand": [
                1000,
                5000,
                10000,
                20000,
                30000,
                40000,
                50000,
                75000,
                100000

            ]
        }



        batch_run = BatchRunnerMP(World,
                                  fixed_parameters=fixed_params,
                                  variable_parameters=variable_params,
                                  iterations=1,
                                  max_steps=number_of_steps, nr_processes=3)

        batch_run.run_all()


if __name__ == "__main__":
    DemandTimer().run_world_with_demand_and_power_plants()



#! /bin/env python3
import os

# import pytz
import os.path
import sys
from datetime import date, datetime, timedelta
# sys.path.append(os.path.join(os.path.dirname(os.path.realpath('__file__')), '../../..'))
# sys.path.insert(0, 'home/alexkell/elecsim3')
# ROOT_DIR = os.path.join(os.path.dirname(os.path.realpath('__file__')), '')
ROOT_DIR = "~/elecsim3/elecsim/"
from joblib import Parallel, delayed
import os.path
import sys
import pickle
import ray
from elecsim.model.world import World

from tqdm import tqdm
import multiprocessing

import sys

import numpy as np
# import multiprocessing
import tracemalloc
from elecsim.constants import ROOT_DIR

import pandas as pd
import pickle
import seaborn as sns
import matplotlib.pyplot as plt
from fitter import Fitter
import fitter
from scipy import stats
import importlib
from scipy.stats import johnsonsb, skewnorm, dgamma, genlogistic, dweibull, johnsonsu


import time
import logging

"""
File name: compare_by_ml_accuracy
Date created: 19/04/2020
Feature: #Enter feature description here
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"

import warnings
warnings.filterwarnings("ignore")


logging.basicConfig(level=logging.INFO)

MARKET_TIME_SPLICES = 8
YEARS_TO_RUN = 25
number_of_steps = YEARS_TO_RUN * MARKET_TIME_SPLICES

# print("ROOT_DIR: {}".format(ROOT_DIR))

scenario_file = "{}/scenario/reference_scenario_beis_2018.py".format(ROOT_DIR)

# scenario_file = "{}/../run/beis_case_study/scenario/reference_scenario_2018.py".format(ROOT_DIR)

# @ray.remote

def run_world(optimal_carbon_tax=None, distribution_name = None, demand_distribution=None):
    beis_params = [0.00121256259168, 46.850377392563864, 0.0029982421515, 28.9229765616468, 0.00106156336814, 18.370337670063762, 0.00228312539654, 0.0, 0.0024046471141100003, 34.43480109190594, 0.0, -20.88014916953091, 0.0, 8.15032953348701, 0.00200271495761, -12.546185375581802, 0.00155518243668, 39.791132970522796, 0.00027449937576, 8.42878689508516, 0.00111989525697, 19.81640207212787, 0.00224091998324, 5.26288570922149, 0.00209189353332, -5.9117317131295195, 0.00240696026847, -5.0144941135222, 0.00021183142492999999, -1.29658413335784, 0.00039441444392000004, -11.41659250225168, 0.00039441444392000004, -11.41659250225168, 120.21276910611674, 0.0, 0.00059945111227]
    # beis_params = [0.00121256259168, 46.850377392563864, 0.0029982421515, 28.9229765616468, 0.00106156336814, 18.370337670063762, 0.00228312539654, 0.0, 0.0024046471141100003, 34.43480109190594, 0.0, -20.88014916953091, 0.0, 8.15032953348701, 0.00200271495761, -12.546185375581802, 0.00155518243668, 39.791132970522796, 0.00027449937576, 8.42878689508516, 0.00111989525697, 19.81640207212787, 0.00224091998324, 5.26288570922149, 0.00209189353332, -5.9117317131295195, 0.00240696026847, -5.0144941135222, 0.00021183142492999999, -1.29658413335784, 0.00039441444392000004, -11.41659250225168, 0.0021988838824299997, 12.633572943294599, 120.21276910611674, 0.0, 0.00059945111227]

    prices_individual = np.array(beis_params[:-3]).reshape(-1, 2).tolist()

    MARKET_TIME_SPLICES = 8
    YEARS_TO_RUN = 25
    number_of_steps = YEARS_TO_RUN * MARKET_TIME_SPLICES

    scenario_2018 = "{}/scenario/reference_scenario_beis_2018.py".format(ROOT_DIR)

    # if individual[0] == 1:
    #     individual = [individual[1]*i + individual[2] for i in range(1, 20)]
    # else:
    #     individual = [individual[3]*i ** individual[4] + individual[2] for i in range(1, 20)]
    # individual = [individual[0]*i + individual[1] for i in range(1, 20)]
    # print(individual)


    world = World(demand_distribution=demand_distribution, distribution_name = distribution_name, carbon_price_scenario=optimal_carbon_tax, initialization_year=2018, scenario_file=scenario_2018, market_time_splices=MARKET_TIME_SPLICES, data_folder="compare_ml_accuracy", number_of_steps=number_of_steps, long_term_fitting_params=prices_individual, highest_demand=63910, nuclear_subsidy=beis_params[-3], future_price_uncertainty_m=beis_params[-2], future_price_uncertainty_c=beis_params[-1])
    for _ in range(YEARS_TO_RUN):
        for i in range(MARKET_TIME_SPLICES):
            # try:
            if i/8 == 0:
                print('end of year')

            world.step()
    # return None


# @ray.remote
# def run_world(num_steps=number_of_steps, demand_distribution=None, long_term_fitting_params=None, optimal_carbon_tax = None):
#     world = World(initialization_year=2018, market_time_splices=MARKET_TIME_SPLICES, data_folder="results", number_of_steps=number_of_steps, scenario_file=scenario_file, demand_distribution=demand_distribution, long_term_fitting_params=long_term_fitting_params, carbon_price_scenario=optimal_carbon_tax)
#     for i in range(num_steps):
#         world.step()


# results = []
if __name__ == '__main__':
    # ray.init(num_cpus=7)

    beis_params = [0.00121256259168, 46.850377392563864, 0.0029982421515, 28.9229765616468, 0.00106156336814, 18.370337670063762, 0.00228312539654, 0.0, 0.0024046471141100003, 34.43480109190594, 0.0, -20.88014916953091, 0.0, 8.15032953348701, 0.00200271495761, -12.546185375581802, 0.00155518243668, 39.791132970522796, 0.00027449937576, 8.42878689508516, 0.00111989525697, 19.81640207212787, 0.00224091998324, 5.26288570922149, 0.00209189353332, -5.9117317131295195, 0.00240696026847, -5.0144941135222, 0.00021183142492999999, -1.29658413335784, 0.00039441444392000004, -11.41659250225168, 0.00039441444392000004, -11.41659250225168, 120.21276910611674, 0.0, 0.00059945111227]

    prices_individual = np.array(beis_params[:-3]).reshape(-1, 2).tolist()

    MARKET_TIME_SPLICES = 8
    YEARS_TO_RUN = 17
    number_of_steps = YEARS_TO_RUN * MARKET_TIME_SPLICES

    scenario_2018 = "reference_scenario_2018.py".format(ROOT_DIR)

    prices_individual = np.array(beis_params[:-3]).reshape(-1, 2).tolist()

    carbon_df = pd.read_csv('linear_data_exploded.csv'.format(ROOT_DIR))
    carbon_list = carbon_df.x.tolist()

    result_distributions_object = pickle.load(open("result_distributions_object.p".format(ROOT_DIR), "rb"))
    # print(result_distributions_object)
    for resultant_dists in result_distributions_object:
        # print(resultant_dists)

        dist_class = eval(list(result_distributions_object[resultant_dists].fitted_param.keys())[0] + ".rvs")
        dist_object = dist_class(*list(result_distributions_object[resultant_dists].fitted_param.values())[0], size=50000).tolist()

        # print(dist_object))
        # run_world(carbon_list, resultant_dists, dist_object)
        Parallel(n_jobs=multiprocessing.cpu_count()-1)(delayed(run_world)(carbon_list, resultant_dists, dist_object) for i in tqdm(range(0, 100)))
        # Parallel(n_jobs=7)(delayed(run_world)(carbon_list, resultant_dists, dist_object) for i in tqdm(range(0, 100)))
        # pool.map(run_world(number_of_steps, dist_object, prices_individual, carbon_list), list(range(0, 150)))

        # time.sleep(30)

    os.execv(sys.executable, [sys.executable] + sys.argv)






import logging
from functools import lru_cache

import numpy as np
import pandas as pd
import os, sys

from elecsim.market.electricity.market.power_exchange import PowerExchange
from elecsim.role.market.latest_market_data import LatestMarketData
from elecsim.role.investment.curve_fitting_functions import logit
from elecsim.scenario.scenario_data import years_for_agents_to_predict_forward
import elecsim.scenario.scenario_data
from scipy.optimize import curve_fit
logger = logging.getLogger(__name__)

"""
File name: predict_load_duration_prices
Date created: 11/01/2019
Feature: # Predict Load Duration Prices
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"


class PredictPriceDurationCurve:

    def __init__(self, model, look_back_period):
        self.model = model
        self.demand_change_predicted = LatestMarketData(self.model).agent_forecast_value("demand", look_back_period, elecsim.scenario.scenario_data.years_for_agents_to_predict_forward)

        # power_ex = PowerExchange(self.model)
        self.power_ex = self.model.PowerExchange

    def predict_price_duration_curve(self):
        self.power_ex.price_duration_curve = []
        if self.model.market_time_splices == 1:
            year_segment_consumption_predicted = [cons * self.demand_change_predicted for cons in self.model.demand.segment_consumption]
            # self.power_ex.tender_bids(self.model.demand.segment_hours, predicted_consumption, predict=True)
            year_segment_hours = self.model.demand.segment_hours
            # self.model.clear_all_bids()
            #
            # predicted_price_duration_curve = power_ex.price_duration_curve
            # predicted_price_duration_curve = estimate_lost_load_price(predicted_price_duration_curve)
        else:
            price_duration_curve = []
            # year_segment_hours, year_segment_consumption = self.model.demand.get_demand_for_year()
            year_segment_hours = self.model.demand.year_segment_hours
            year_segment_consumption = self.model.demand.year_segment_consumption

            # for segment_hours, segment_consumption in zip(year_segment_hours, year_segment_consumption):

            year_segment_consumption_predicted = [cons * self.demand_change_predicted for cons in year_segment_consumption]

        if elecsim.scenario.scenario_data.investment_mechanism == "future_price_fit":
            predicted_price_duration_curve = self.fit_linear_price_duration_curve(year_segment_consumption_predicted, year_segment_hours)
        elif elecsim.scenario.scenario_data.investment_mechanism == "projection_fit":
            predicted_price_duration_curve = self.fit_linear_price_duration_curve_from_projections(year_segment_consumption_predicted, year_segment_hours)
            predicted_price_duration_curve_market = self.market_price_curve_prediction(year_segment_consumption_predicted, year_segment_hours)
            linear_price_mean = predicted_price_duration_curve['accepted_price'].mean()
            market_price_mean = predicted_price_duration_curve_market['accepted_price'].mean()
            if market_price_mean < linear_price_mean:
                predicted_price_duration_curve = predicted_price_duration_curve_market
        else:
            predicted_price_duration_curve = self.market_price_curve_prediction(year_segment_consumption_predicted, year_segment_hours)

        logger.debug("predicted_price_duration_curve: {}".format(predicted_price_duration_curve))
        # predicted_price_duration_curve.segment_hour = predicted_price_duration_curve.segment_hour.cumsum()
        # price_duration_curve.append(predicted_price_duration_curve_day)
        # predicted_price_duration_curve = pd.concat(price_duration_curve)
        # predicted_price_duration_curve.sort_values("segment_demand", ascending=False, inplace=True)

        return predicted_price_duration_curve

    def market_price_curve_prediction(self, year_segment_consumption_predicted, year_segment_hours):
        self.power_ex.tender_bids(year_segment_hours, year_segment_consumption_predicted, predict=True)
        self.model.clear_all_bids()
        predicted_price_duration_curve = self.power_ex.price_duration_curve
        predicted_price_duration_curve = estimate_lost_load_price(predicted_price_duration_curve)
        return predicted_price_duration_curve

    def fit_linear_price_duration_curve(self, year_segment_consumption_predicted, year_segment_hours):
        segment_demand = year_segment_consumption_predicted
        segment_hour = year_segment_hours
        year = self.model.year_number

        if self.model.fitting_params is not None:
            fitting_params = self.model.fitting_params
        elif self.model.long_term_fitting_params is not None:
            fitting_params = self.model.long_term_fitting_params[self.model.years_from_start]

            holder_params = fitting_params.copy()


        # if self.model.fitting_params is not None and self.model.long_term_fitting_params is not None:
        if hasattr(self.model, "fitting_params") or hasattr(self.model, "long_term_fitting_params"):
            try:
                holder_params[0] *= np.random.normal(0, self.model.future_price_uncertainty_m)
                holder_params[1] *= np.random.normal(0, self.model.future_price_uncertainty_c)
            except:
                pass

        temp_price_dataframe = {"segment_demand": segment_demand, "segment_hour": segment_hour, "year": year}

        price_duration_curve = pd.DataFrame(temp_price_dataframe)
        price_duration_curve['accepted_price'] = fitting_params[0]*price_duration_curve.segment_demand + fitting_params[1]
        # price_duration_curve['accepted_price'] = fitting_params[0]*segment_hour + fitting_params[1]

        predicted_price_duration_curve = price_duration_curve
        return predicted_price_duration_curve

    def fit_linear_price_duration_curve_from_projections(self, year_segment_consumption_predicted, year_segment_hours):
        segment_demand = year_segment_consumption_predicted
        segment_hour = year_segment_hours
        year = self.model.year_number + years_for_agents_to_predict_forward
        year_for_prediction = self.model.years_from_start + years_for_agents_to_predict_forward
        try:
            baseload_price = electricity_baseload[year_for_prediction]
            volume_weighted_price = electricity_volume_weighted[year_for_prediction]
        except:
            baseload_price = electricity_baseload[-1]
            volume_weighted_price = electricity_volume_weighted[-1]

        temp_price_dataframe = {"segment_demand": segment_demand, "segment_hour": segment_hour, "accepted_price": np.NaN, "year": year}

        price_duration_curve = pd.DataFrame(temp_price_dataframe)
        # price_duration_curve.accepted_price =
        price_duration_curve.accepted_price.iloc[len(price_duration_curve)-1] = baseload_price
        price_duration_curve.accepted_price.iloc[int((len(price_duration_curve)-1)/2)] = volume_weighted_price
        price_duration_curve.accepted_price = price_duration_curve.accepted_price.interpolate(method="spline", order=1, limit_direction="both")


        predicted_price_duration_curve = price_duration_curve
        return predicted_price_duration_curve


def estimate_lost_load_price(predicted_price_duration_curve):
    if elecsim.scenario.scenario_data.lost_load_price_predictor:
        if all(predicted_price_duration_curve.accepted_price == elecsim.scenario.scenario_data.lost_load):
            return predicted_price_duration_curve
        if any(predicted_price_duration_curve.accepted_price == elecsim.scenario.scenario_data.lost_load):
            predicted_price_duration_curve.accepted_price = predicted_price_duration_curve.accepted_price.replace(elecsim.scenario.scenario_data.lost_load,np.nan)
            # predicted_price_duration_curve.accepted_price.interpolate(method="polynomial", order=1)
            if predicted_price_duration_curve.accepted_price.count() > 1:

                predicted_price_duration_curve_training = predicted_price_duration_curve.dropna()

                logger.debug("predicted_price_duration_curve: {}".format(predicted_price_duration_curve))

                # popt, pcov = curve_fit(logit, predicted_price_duration_curve_training.segment_demand, predicted_price_duration_curve_training.accepted_price, [1, 1, 1, 1, 90])
                # extrapolated = logit(predicted_price_duration_curve.loc[np.isnan(predicted_price_duration_curve.accepted_price), 'segment_demand'], *popt)

                # # Polynomial regression fitting
                p = np.poly1d(np.polyfit(predicted_price_duration_curve_training.segment_demand, predicted_price_duration_curve_training.accepted_price,3))
                extrapolated = p(predicted_price_duration_curve.loc[np.isnan(predicted_price_duration_curve.accepted_price), 'segment_demand'])
                # logger.debug("extrapolated: {}".format(extrapolated))
                predicted_price_duration_curve.loc[np.isnan(predicted_price_duration_curve.accepted_price), "accepted_price"] = extrapolated
                return predicted_price_duration_curve
            elif predicted_price_duration_curve.accepted_price.count() == 1:
                predicted_price_duration_curve.accepted_price = predicted_price_duration_curve.accepted_price.fillna(predicted_price_duration_curve.accepted_price.mean())
                return predicted_price_duration_curve
        else:
            return predicted_price_duration_curve
    else:
        return predicted_price_duration_curve


# @lru_cache(1024)
def get_price_duration_curve(model, look_back_period):
    predicted_price_duration_curve = PredictPriceDurationCurve(model, look_back_period=look_back_period).predict_price_duration_curve()
    return predicted_price_duration_curve




import pandas as pd
pd.set_option('display.max_rows', 500)
pd.set_option('display.max_columns', 500)
from elecsim.constants import ROOT_DIR, KW_TO_MW
import numpy as np

from scipy.optimize import root

KTOE_TO_MWH = 11630

investment_mechanism = "future_price_fit"
# investment_mechanism = "projection_fit"
potential_plants_to_invest = ['CCGT H Class', 'CCGT F Class', 'CCGT CHP mode', 'Coal - CCS ASC NH3 FOAK', 'Coal - CCS ASC Partial FOAK', 'Coal - CCS ASC FOAK', 'Nuclear - PWR FOAK', 'Onshore UK>5MW', 'Offshore R2', 'Offshore R3', 'PV>5MW', 'Recip Gas 2000 hr', 'RECIP GAS 500 hrs', 'Recip Diesel 2000 hr', 'Recip Diesel 500 hrs', 'Recip Diesel 90 hrs']

multi_year_data = pd.read_csv('{}/data/processed/multi_day_data/4_medoids.csv'.format(ROOT_DIR))
multi_year_data_scaled = pd.read_csv('{}/data/processed/multi_day_data/4_medoids_scaled.csv'.format(ROOT_DIR))


# BEIS Final Electricity Consumption
electricity_ktoe = [24822.84204, 24470.32991, 24301.10466, 24288.25851, 24438.20398, 24717.02662, 24995.8011, 25361.32776, 25813.71342, 26342.63171, 26893.41806, 27471.23698, 27867.29557, 28262.51807, 28754.44771, 29191.80642, 29648.75735]
renewables_ktoe = [7609.872531, 8007.503459, 8179.451416, 8267.036507, 8314.81648, 8323.254324, 8317.588187, 8318.702017, 8325.651362, 8328.676807, 8335.161717, 8333.378709, 8327.276614, 8303.48562, 8198.764019, 7988.909411, 7737.627294]

final_consumption = [(elec+renew) * 11630 for elec, renew in zip(electricity_ktoe, renewables_ktoe)]

def get_difference(scaler, df, required_mwh):
    load_dat = df[df.data_type=='load']
    load_total = load_dat.capacity_factor * scaler
    return load_total.sum() - required_mwh

demand_sizes = [root(fun=get_difference, x0=50000, args=(multi_year_data_scaled, consumption)).x for consumption in final_consumption]
demand_sizes_repeated = np.repeat(demand_sizes, 8).tolist()


# Demand per segment of load duration function
segment_demand_diff = [17568, 21964, 23127, 24327, 25520, 26760, 27888, 28935, 29865, 30721, 31567, 32315, 33188, 34182, 35505, 37480, 39585, 42206, 45209, 52152]
segment_demand = [52152, 45209, 42206, 39585, 37480, 35505, 34182, 33188, 32315, 31567, 30721, 29865, 28935, 27888, 26760, 25520, 24327, 23127, 21964, 17568]
# Time of load duration function
segment_time = [8752.5, 8291.83, 7831.17, 7370.5, 6909.92, 6449.25, 5988.58, 5527.92, 5067.25, 4606.58, 4146, 3685.33, 3224.67, 2764, 2303.33, 1842.67, 1382.08, 921.42, 460.75, 0.08]

# Change in load duration function by year
# yearly_demand_change = [0.949620, 0.959511, 0.979181, 0.984285, 0.987209, 0.983118]

yearly_demand_change = [1-((b[0] - a[0]) / a[0]) for a, b in zip(demand_sizes[::1], demand_sizes[1::1])]


# First year maximum demand size
initial_max_demand_size = demand_sizes[0]

# Electricity Prices
electricity_volume_weighted = [58, 56, 53, 52, 53, 54, 57, 58, 58, 60, 58, 58, 59, 61, 60, 63, 60, 58]
electricity_baseload = [58, 55, 52, 51, 53, 53, 56, 57, 57, 59, 57, 57, 58, 59, 59, 61, 58, 56]

multi_year_data = pd.read_csv('{}/data/processed/multi_day_data/4_medoids.csv'.format(ROOT_DIR))
multi_year_data_scaled = pd.read_csv('{}/data/processed/multi_day_data/4_medoids_scaled.csv'.format(ROOT_DIR))
# Fuel prices (£/MWh)

# Historical fuel prices of coal, oil and gas  Source: Average prices of fuels purchased by the major UK power producers, BEIS UK Government, table_321.xlsx
historical_fuel_prices_long = pd.read_csv('{}/data/processed/fuel/fuel_costs/historical_fuel_costs/historical_fuel_costs_converted_long.csv'.format(ROOT_DIR))
historical_fuel_prices_mw = pd.read_csv('{}/data/processed/fuel/fuel_costs/historical_fuel_costs/fuel_costs_per_mwh.csv'.format(ROOT_DIR))


# Future $/GBP exchange rate
dollar_gbp_exchange_rate = [1.36, 1.38, 1.40, 1.40, 1.40, 1.40, 1.40, 1.40, 1.40, 1.40, 1.40, 1.40, 1.40, 1.40, 1.40, 1.40, 1.40]

# Future fuel prices
gas_scenario = [53.0, 48.0, 49.0, 51.0, 52.0, 54.0, 56.0, 57.0, 59.0, 60.0, 62.0, 63.0, 63.0, 63.0, 63.0, 63.0, 63.0]
gas_scenario = [price / 0.0293001 / 100 for price in gas_scenario]
# gas_price = np.repeat(gas_scenario, 8).tolist()
gas_price = gas_scenario

coal_scenario = [85.7, 85.7, 85.7, 85.7, 85.7, 85.7, 85.7, 86.7, 86.7, 86.7, 86.7, 86.7, 86.7, 86.7, 86.7, 86.7, 86.7]
coal_price = [price/8.141/exchange  for price, exchange in zip(coal_scenario, dollar_gbp_exchange_rate)]
# coal_price = np.repeat(coal_scenario, 8).tolist()

oil_scenario = [70.7, 71.7, 72.7, 74.7, 75.7, 76.7, 77.7, 79.7, 80.7, 81.7, 83.7, 84.7, 84.7, 84.7, 84.7, 84.7, 84.7]
oil_scenario = [price/1.69941/exchange for price, exchange in zip(oil_scenario, dollar_gbp_exchange_rate)]

# gas_price = [KW_TO_MW * 0.01909] * 60  # Source: Average prices of fuels purchased by the major UK power producers: table_321.xls
# coal_price = [KW_TO_MW * 0.01106] * 60  # Source: Average prices of fuels purchased by the major UK power producers: table_321.xls
uranium_price = [KW_TO_MW * 0.0039] * 17  # Source: The Economics of Nuclear Power: EconomicsNP.pdf
oil_price = [KW_TO_MW * 0.02748] * 17  # Source: Average prices of fuels purchased by the major UK power producers: table_321.xls
diesel_price = [KW_TO_MW * 0.1] * 17  # Source: https://www.racfoundation.org/data/wholesale-fuel-prices-v-pump-prices-data
woodchip_price = [KW_TO_MW * 0.0252] * 17  # Source: Biomass for Power Generation: IRENA BiomassCost.pdf
poultry_litter_price = [KW_TO_MW * 0.01139] * 17  # Source: How much is poultry litter worth?: sp06ca08.pdf
straw_price = [KW_TO_MW * 0.016488] * 17  # Source: https://dairy.ahdb.org.uk/market-information/farm-expenses/hay-straw-prices/#.W6JnFJNKiYU
meat_price = [KW_TO_MW * 0.01] * 17  # Assumption: Low price due to plant_type being a waste product
waste_price_post_2000 = [KW_TO_MW * -0.0252] * 17  # Source: Gate fees report 2017 Comparing the costs of waste treatment options: Gate Fees report 2017_FINAL_clean.pdf
waste_price_pre_2000 = [KW_TO_MW * -0.01551] * 17  # Source: Gate fees report 2017 Comparing the costs of waste treatment options: Gate Fees report 2017_FINAL_clean.pdf

# Joining historical and future fuel prices for simulation purposes.
fuel_prices = pd.DataFrame(data=[coal_price, oil_price, gas_price, uranium_price, diesel_price, woodchip_price,
                                 poultry_litter_price, straw_price, meat_price, waste_price_post_2000,
                                 waste_price_pre_2000],
                           columns=[str(i) for i in range(2019, (2019+len(gas_price)))])

fuel_prices = pd.concat([historical_fuel_prices_mw, fuel_prices], axis=1)
# Convert from wide to long
fuel_prices = fuel_prices.melt(id_vars=['Fuel'], var_name='Year', value_vars=list(fuel_prices.loc[:,'1990':'2035'].columns))
fuel_prices.Year = pd.to_numeric(fuel_prices.Year)
# Fill NA's with average of group
fuel_prices['value'] = fuel_prices.groupby("Fuel")['value'].transform(lambda x: x.fillna(x.mean()))


# Weighted Cost of Capital
nuclear_wacc = 0.1  # https://www.imperial.ac.uk/media/imperial-college/research-centres-and-groups/icept/Cost-estimates-for-nuclear-power-in-the-UK.pdf (page 20). # Post tax
non_nuclear_wacc = 0.059  # https://assets.kpmg/content/dam/kpmg/ch/pdf/cost-of-capital-study-2017-en.pdf # post tax

# Availability
non_fuel_plant_availability = 0.97
pv_availability = 0.995  # https://ieeexplore.ieee.org/document/7355976
offshore_availability = 0.95  # https://pureportal.strath.ac.uk/files-asset/43185998/Carroll_etal_EWEA2015_availability_improvements_from_condition_monitoring_systems.pdf
onshore_availability = 0.97  # https://pureportal.strath.ac.uk/files-asset/43185998/Carroll_etal_EWEA2015_availability_improvements_from_condition_monitoring_systems.pdf
fuel_plant_availability = 0.97  # Electricity Generation Costs and Hurdle Rates - Leigh_Fisher_Non-renewable_Generation_Cost.pdf


# Capacity factor data (from https://www.renewables.ninja/)
# Wind
wind_capacity_factor = pd.read_csv('{}/data/processed/capacity_factor/Wind/ninja_wind_country_GB_current-merra-2_corrected.csv'.format(ROOT_DIR))
# wind_capacity_factor.time = pd.to_datetime(wind_capacity_factor.time)
# Solar
solar_capacity_factor = pd.read_csv('{}/data/processed/capacity_factor/Solar/ninja_pv_country_GB_merra-2_corrected.csv'.format(ROOT_DIR))
# Hydro
hydro_capacity_factor = 0.456 # http://www.osemosys.org/uploads/1/8/5/0/18504136/hydropower.pdf
# Nuclear
nuclear_capacity_factor = 0.92 * 0.826  # https://www.energy.gov/ne/articles/what-generation-capacity and https://pris.iaea.org/PRIS/WorldStatistics/ThreeYrsEnergyAvailabilityFactor.aspx

# Availability factors (from Source: AESO 2017 Annual Market Statistics)
historical_availability_factor = pd.read_csv('{}/data/processed/availability_factor/historical_availability_factor.csv'.format(ROOT_DIR))

# UK Hourly Demand
historical_hourly_demand = pd.read_csv('{}/data/processed/electricity_demand/uk_all_year_demand.csv'.format(ROOT_DIR))
# historical_hourly_demand = historical_hourly_demand.tail(2628)

# Load duration curve:
load_duration_curve = pd.read_csv('{}/data/processed/load_duration_curve/load_duration_curve.csv'.format(ROOT_DIR))
load_duration_curve_diff = pd.read_csv('{}/data/processed/load_duration_curve/load_duration_curve_difference.csv'.format(ROOT_DIR))

# Learning rate for renewables
learning_rate = 0.5

# Generator Companies imported from Government data files
power_plants = pd.read_csv('{}/data/processed/power_plants/uk_power_plants/uk_power_plants.csv'.format(ROOT_DIR), dtype={'Start_date': int})
# power_plants = power_plants[:100]
modern_plant_costs = pd.read_csv('{}/data/processed/power_plants/power_plant_costs/modern_power_plant_costs/power_plant_costs_with_simplified_type.csv'.format(ROOT_DIR))

power_plant_historical_costs_long = pd.read_csv('{}/data/processed/power_plants/power_plant_costs/historical_power_plant_costs/historical_power_plant_costs_long.csv'.format(ROOT_DIR))

# Variable operation and maintenance costs random numbers for stochasticity (uniform distribution)
o_and_m_multiplier = (0.3, 2)

# Historical power plant efficiency
historical_fuel_plant_efficiency = pd.read_csv('{}/data/processed/power_plants/power_plant_costs/historical_power_plant_costs/efficiency/historical_fuel_plant_efficiency.csv'.format(ROOT_DIR))  # https://www.eia.gov/electricity/annual/html/epa_08_01.html, U.S. Energy Information Administration, Form EIA-923, "Power Plant Operations Report," and predecessor form(s) including U.S. Energy Information Administration, Form EIA-906, "Power Plant Report;" and Form EIA-920, "Combined Heat and Power Plant Report;" Form EIA-860, "Annual Electric Generator Report."


# Company financials
company_financials = pd.read_csv('{}/data/processed/companies/company_financials.csv'.format(ROOT_DIR))

# Bid mark-up price
bid_mark_up = 1.0


# Carbon price - Forecast used from BEIS Electricity Generation Report - Page 10 - Includes forecast for carbon tax and EU ETS
# carbon_price_scenario = [18.00, 19.42, 20.83, 22.25, 23.67, 25.08, 26.50, 27.92, 29.33, 30.75, 32.17, 33.58, 35.00, 43.25, 51.50, 59.75, 68.00, 76.25, 84.50, 92.75, 101.00, 109.25, 117.50, 125.75, 134.00, 142.25, 150.50, 158.75, 167.00, 175.25, 183.50, 191.75, 200.00]
# carbon_price_scenario = [10]*1000
carbon_price_scenario = [30.8, 30.9, 31.7, 31.9, 32.3, 32.6, 32.9, 33.2, 33.5, 33.9, 34.2, 35.6, 42.7, 54.0, 65.3, 76.6, 87.9, 99.3]
# carbon_price_scenario = np.repeat(carbon_scenario, 8).tolist()
# carbon_price_scenario = [0]*100
# EU_ETS_COST = 13.62
# carbon_price_scenario = [uk_tax + EU_ETS_COST for uk_tax in carbon_price_scenario]


def concatenate_carbon_price():
      carbon_data = {'year': [str(i) for i in range(2019, (2019 + len(carbon_price_scenario)))],
                    'price': carbon_price_scenario}
      carbon_price_scenario_df = pd.DataFrame(carbon_data)
      historical_carbon_price = pd.read_csv(ROOT_DIR + '/data/processed/carbon_price/uk_carbon_tax_historical.csv')
      carbon_cost = historical_carbon_price.append(carbon_price_scenario_df, sort=True)
      carbon_cost.year = pd.to_numeric(carbon_cost.year)
      return carbon_cost


carbon_price_all_years = concatenate_carbon_price()



# Join historical and future carbon prices into dataframe for simulation purposes
# carbon_data = {'year': [str(i) for i in range(2019, (2019 + len(carbon_price_scenario)))], 'price': carbon_price_scenario}
# carbon_price_scenario_df = pd.DataFrame(carbon_data)
# historical_carbon_price = pd.read_csv(ROOT_DIR + '/data/processed/carbon_price/uk_carbon_tax_historical.csv')
# carbon_cost = historical_carbon_price.append(carbon_price_scenario_df, sort=True)
# carbon_cost.year = pd.to_numeric(carbon_cost.year)

# Lost load price - Set at £6000 MW/h as per the recommendations of the UK Government https://assets.publishing.service.gov.uk/government/uploads/system/uploads/attachment_data/file/267613/Annex_C_-_reliability_standard_methodology.pdf
# lost_load = 6000
lost_load = 300

upfront_investment_costs = 0.25

years_for_agents_to_predict_forward = 10


max_onshore_capacity = 34000  # 42 million acres (170,000 km^2) of agricultural land in UK, 10% dedicated to wind with average energy density of 2W/m^2 David McKay, http://www.withouthotair.com/cft.pdf),
max_offshore_capacity = 1  #
max_pv_capacity = 85000  # 10W/m^2 (David McKay), cover 5% of the UK

# Predict lost load through exponential and linear regression
lost_load_price_predictor = True

known_plant_retirements = {}



import sys
# import scenario_file
# sys.modules['elecsim'].scenario.scenario_data=scenario_file

from elecsim.model.world import World

import logging
# logging.basicConfig(level=logging.INFO)


if __name__ == "__main__":
    world = World(2018, log_level="info", market_time_splices=8)
    for i in range(20):
        world.step()
